from typing import Any, Callable, ClassVar, Literal

from pydantic import Field, field_validator, model_validator
from typing_extensions import Self

from polaris.benchmark import BenchmarkSpecification
from polaris.evaluate.utils import evaluate_benchmark
from polaris.utils.types import IncomingPredictionsType

from polaris.evaluate import BenchmarkResultsV2
from polaris.benchmark._split_v2 import SplitSpecificationV2Mixin
from polaris.dataset import DatasetV2, Subset
from polaris.utils.errors import InvalidBenchmarkError
from polaris.utils.types import ColumnName


class BenchmarkV2Specification(SplitSpecificationV2Mixin, BenchmarkSpecification[BenchmarkResultsV2]):
    _version: ClassVar[Literal[2]] = 2

    dataset: DatasetV2 = Field(exclude=True)
    n_classes: dict[ColumnName, int] = Field(default_factory=dict)

    @field_validator("dataset", mode="before")
    @classmethod
    def _parse_dataset(
        cls,
        v: DatasetV2 | str | dict[str, Any],
    ) -> DatasetV2:
        """
        Allows either passing a Dataset object or the kwargs to create one
        """
        match v:
            case dict():
                return DatasetV2(**v)
            case str():
                return DatasetV2.from_json(v)
            case DatasetV2():
                return v

    @model_validator(mode="after")
    def _validate_n_classes(self) -> Self:
        """
        The number of classes for each of the target columns.
        """
        columns = set(self.n_classes.keys())
        if not columns.issubset(self.target_cols):
            raise InvalidBenchmarkError("Not all specified class numbers were found in the target columns.")

        return self

    @model_validator(mode="after")
    def _validate_split_in_dataset(self) -> Self:
        """
        Verifies that:
          - All indices are valid given the dataset
        """
        dataset_length = len(self.dataset)
        if self.split.max_index >= dataset_length:
            raise InvalidBenchmarkError("The predefined split contains invalid indices")

        return self

    def _get_test_sets(
        self, hide_targets=True, featurization_fn: Callable | None = None
    ) -> dict[str, Subset]:
        """
        Construct the test set(s), given the split in the benchmark specification. Used
        internally to construct the test set for client use and evaluation.
        """
        # TODO: We need a subset class that can handle very large index sets without copying or materializing all of them
        return {
            label: self._get_subset(index_set.indices, hide_targets, featurization_fn)
            for label, index_set in self.split.test_items()
        }

    def get_train_test_split(
        self, featurization_fn: Callable | None = None
    ) -> tuple[Subset, dict[str, Subset]]:
        """Construct the train and test sets, given the split in the benchmark specification.

        Returns [`Subset`][polaris.dataset.Subset] objects, which offer several ways of accessing the data
        and can thus easily serve as a basis to build framework-specific (e.g. PyTorch, Tensorflow)
        data-loaders on top of.

        Args:
            featurization_fn: A function to apply to the input data. If a multi-input benchmark, this function
                expects an input in the format specified by the `input_format` parameter.

        Returns:
            A tuple with the train `Subset` and test `Subset` objects.
                If there are multiple test sets, these are returned in a dictionary and each test set has
                an associated name. The targets of the test set can not be accessed.
        """
        train = self._get_subset(
            self.split.training.indices, hide_targets=False, featurization_fn=featurization_fn
        )
        test = self._get_test_sets(hide_targets=True, featurization_fn=featurization_fn)
        return train, test

    def evaluate(
        self,
        y_pred: IncomingPredictionsType | None = None,
        y_prob: IncomingPredictionsType | None = None,
    ) -> BenchmarkResultsV2:
        """Execute the evaluation protocol for the benchmark, given a set of predictions.

        info: What about `y_true`?
            Contrary to other frameworks that you might be familiar with, we opted for a signature that includes just
            the predictions. This reduces the chance of accidentally using the test targets during training.

        For this method, we make the following assumptions:

        1. There can be one or multiple test set(s);
        2. There can be one or multiple target(s);
        3. The metrics are _constant_ across test sets;
        4. The metrics are _constant_ across targets;
        5. There can be metrics which measure across tasks.

        Args:
            y_pred: The predictions for the test set, as NumPy arrays.
                If there are multiple targets, the predictions should be wrapped in a dictionary with the target labels as keys.
                If there are multiple test sets, the predictions should be further wrapped in a dictionary
                    with the test subset labels as keys.
            y_prob: The predicted probabilities for the test set, formatted similarly to predictions, based on the
                number of tasks and test sets.

        Returns:
            A `BenchmarkResultsV2` object. This object can be directly submitted to the Polaris Hub.

        Examples:
            1. For regression benchmarks:
                pred_scores = your_model.predict_score(molecules) # predict continuous score values
                benchmark.evaluate(y_pred=pred_scores)
            2. For classification benchmarks:
                - If `roc_auc` and `pr_auc` are in the metric list, both class probabilities and label predictions are required:
                    pred_probs = your_model.predict_proba(molecules) # predict probablities
                    pred_labels = your_model.predict_labels(molecules) # predict class labels
                    benchmark.evaluate(y_pred=pred_labels, y_prob=pred_probs)
                - Otherwise:
                    benchmark.evaluate(y_pred=pred_labels)
        """

        # Instead of having the user pass the ground truth, we extract it from the benchmark spec ourselves.
        y_true = self._get_test_sets(hide_targets=False)

        scores = evaluate_benchmark(
            target_cols=list(self.target_cols),
            test_set_labels=self.test_set_labels,
            test_set_sizes=self.test_set_sizes,
            metrics=self.metrics,
            y_true=y_true,
            y_pred=y_pred,
            y_prob=y_prob,
        )

        return BenchmarkResultsV2(results=scores, benchmark_artifact_id=self.artifact_id)



import abc
import json
import logging
from os import PathLike
from pathlib import Path, PurePath
from typing import Any, Iterable, MutableMapping
from uuid import uuid4

import fsspec
import numpy as np
import zarr
from pydantic import (
    Field,
    PrivateAttr,
    computed_field,
    field_serializer,
    field_validator,
    model_validator,
)
from typing_extensions import Self

from polaris._artifact import BaseArtifactModel
from polaris.dataset._adapters import Adapter
from polaris.dataset._column import ColumnAnnotation
from polaris.dataset.zarr import MemoryMappedDirectoryStore
from polaris.dataset.zarr._utils import check_zarr_codecs, load_zarr_group_to_memory
from polaris.utils.constants import DEFAULT_CACHE_DIR
from polaris.utils.context import track_progress
from polaris.utils.dict2html import dict2html
from polaris.utils.errors import InvalidDatasetError
from polaris.utils.types import (
    AccessType,
    ChecksumStrategy,
    DatasetIndex,
    HttpUrlString,
    HubOwner,
    SupportedLicenseType,
    ZarrConflictResolution,
)

logger = logging.getLogger(__name__)

# Constants
_CACHE_SUBDIR = "datasets"


class BaseDataset(BaseArtifactModel, abc.ABC):
    """Base data-model for a Polaris dataset, implemented as a [Pydantic](https://docs.pydantic.dev/latest/) model.

    At its core, a dataset in Polaris can _conceptually_ be thought of as tabular data structure that stores data-points
    in a row-wise manner, where each column correspond to a variable associated with that datapoint.

    A Dataset can have multiple modalities or targets, can be sparse and can be part of one or multiple benchmarks.

    Attributes:
        default_adapters: The adapters that the Dataset recommends to use by default to change the format of the data
            for specific columns.
        zarr_root_path: The data for any pointer column should be saved in the Zarr archive this path points to.
        readme: Markdown text that can be used to provide a formatted description of the dataset.
            If using the Polaris Hub, it is worth noting that this field is more easily edited through the Hub UI
            as it provides a rich text editor for writing markdown.
        annotations: Each column _can be_ annotated with a [`ColumnAnnotation`][polaris.dataset.ColumnAnnotation] object.
            Importantly, this is used to annotate whether a column is a pointer column.
        source: The data source, e.g. a DOI, Github repo or URI.
        license: The dataset license. Polaris only supports some Creative Commons licenses. See [`SupportedLicenseType`][polaris.utils.types.SupportedLicenseType] for accepted ID values.
        curation_reference: A reference to the curation process, e.g. a DOI, Github repo or URI.
        artifact_version: The version of the dataset.
        artifact_changelog: A description of the changes made in this dataset version.

    For additional metadata attributes, see the base classes.

    Raises:
        InvalidDatasetError: If the dataset does not conform to the Pydantic data-model specification.
    """

    # Public attributes
    # Data
    default_adapters: dict[str, Adapter] = Field(default_factory=dict)
    zarr_root_path: str | None = None

    # Additional metadata
    readme: str = ""
    annotations: dict[str, ColumnAnnotation] = Field(default_factory=dict)
    source: HttpUrlString | None = None
    license: SupportedLicenseType | None = None
    curation_reference: HttpUrlString | None = None

    # Version-related fields
    artifact_version: int = Field(default=1, frozen=True)
    artifact_changelog: str | None = None

    # Private attributes
    _zarr_root: zarr.Group | None = PrivateAttr(None)
    _zarr_data: MutableMapping[str, np.ndarray] | None = PrivateAttr(None)
    _warn_about_remote_zarr: bool = PrivateAttr(True)
    _cache_dir: str | None = PrivateAttr(None)  # Where to cache the data to if cache() is called.
    _verify_checksum_strategy: ChecksumStrategy = PrivateAttr("verify_unless_zarr")

    @field_validator("default_adapters", mode="before")
    def _validate_adapters(cls, value):
        """Validate the adapters"""
        return {k: Adapter[v] if isinstance(v, str) else v for k, v in value.items()}

    @field_serializer("default_adapters")
    def _serialize_adapters(self, value: dict[str, Adapter]) -> dict[str, str]:
        """Serializes the adapters"""
        return {k: v.name for k, v in value.items()}

    @model_validator(mode="after")
    def _validate_base_dataset_model(self) -> Self:
        # Verify that all annotations are for columns that exist
        if any(k not in self.columns for k in self.annotations):
            raise InvalidDatasetError(
                f"There are annotations for columns that do not exist. Columns: {self.columns}. Annotations: {list(self.annotations.keys())}"
            )

        # Verify that all adapters are for columns that exist
        if any(k not in self.columns for k in self.default_adapters.keys()):
            raise InvalidDatasetError(
                f"There are default adapters for columns that do not exist. Columns: {self.columns}. Adapters: {list(self.annotations.keys())}"
            )

        # Set a default for missing annotations and convert strings to Modality
        for c in self.columns:
            if c not in self.annotations:
                self.annotations[c] = ColumnAnnotation()
            self.annotations[c].dtype = self.dtypes[c]

        return self

    @model_validator(mode="after")
    def _ensure_cache_dir_exists(self) -> Self:
        """
        Set the default cache dir if none and make sure it exists
        """
        if self._cache_dir is None:
            self._cache_dir = str(PurePath(DEFAULT_CACHE_DIR) / _CACHE_SUBDIR / str(uuid4()))
        fs, path = fsspec.url_to_fs(self._cache_dir)
        fs.mkdirs(path, exist_ok=True)

        return self

    @property
    def uses_zarr(self) -> bool:
        """Whether any of the data in this dataset is stored in a Zarr Archive."""
        return self.zarr_root_path is not None

    @property
    def zarr_data(self):
        """Get the Zarr data.

        This is different from the Zarr Root, because to optimize the efficiency of
        data loading, a user can choose to load the data into memory as a numpy array

        Note: General purpose dataloader.
            The goal with Polaris is to provide general purpose datasets that serve as good
            options for a _wide variety of use cases_. This also implies you should be able to
            optimize things further for a specific use case if needed.
        """
        if self._zarr_data is not None:
            return self._zarr_data
        return self.zarr_root

    @property
    def zarr_root(self) -> zarr.Group | None:
        """Get the zarr Group object corresponding to the root.

        Opens the zarr archive in read-write mode if it is not already open.

        Note: Different to `zarr_data`
            The `zarr_data` attribute references either to the Zarr archive or to a in-memory copy of the data.
            See also `dataset.load_to_memory()`.
        """

        from polaris.hub.storage import StorageSession

        if self._zarr_root is not None:
            return self._zarr_root

        if self.zarr_root_path is None:
            return None

        saved_on_hub = self.zarr_root_path.startswith(StorageSession.polaris_protocol)

        if self._warn_about_remote_zarr and saved_on_hub:
            # TODO (cwognum): The user now has no easy way of knowing whether the dataset is "small enough".
            logger.warning(
                f"You're loading data from a remote location. "
                f"If the dataset is small enough, consider caching the dataset first "
                f"using {self.__class__.__name__}.cache() for more performant data access."
            )
            self._warn_about_remote_zarr = False

        try:
            if saved_on_hub:
                self._zarr_root = self.load_zarr_root_from_hub()
            else:
                self._zarr_root = self.load_zarr_root_from_local()

        except KeyError as error:
            raise InvalidDatasetError(
                "A Zarr archive associated with a Polaris dataset has to be consolidated."
            ) from error

        check_zarr_codecs(self._zarr_root)
        return self._zarr_root

    @abc.abstractmethod
    def load_zarr_root_from_hub(self):
        """
        Loads a Zarr archive from the Polaris Hub.
        """
        raise NotImplementedError

    def load_zarr_root_from_local(self):
        """
        Loads a locally stored Zarr archive.

        We use memory mapping by default because our experiments show that it's consistently faster
        """
        if not Path(self.zarr_root_path).exists():
            raise FileNotFoundError(
                f"The Zarr archive at {self.zarr_root_path} does not exist. "
                "If the data is not stored on the Polaris Hub, it needs to be stored locally."
            )
        store = MemoryMappedDirectoryStore(self.zarr_root_path)
        return zarr.open_consolidated(store, mode="r+")

    @computed_field
    @property
    def n_rows(self) -> int:
        """The number of rows in the dataset."""
        return len(self.rows)

    @computed_field
    @property
    def n_columns(self) -> int:
        """The number of columns in the dataset."""
        return len(self.columns)

    @property
    @abc.abstractmethod
    def rows(self) -> Iterable[str | int]:
        """Return all row indices for the dataset"""
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def columns(self) -> list[str]:
        """Return all columns for the dataset"""
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def dtypes(self) -> dict[str, np.dtype]:
        """Return the dtype for each of the columns for the dataset"""
        raise NotImplementedError

    @abc.abstractmethod
    def should_verify_checksum(self, strategy: ChecksumStrategy) -> bool:
        raise NotImplementedError

    def load_to_memory(self):
        """
        Load data from zarr files to memeory

        Warning: Make sure the uncompressed dataset fits in-memory.
            This method will load the **uncompressed** dataset into memory. Make
            sure you actually have enough memory to store the dataset.
        """
        data = self.zarr_data

        if not isinstance(data, zarr.Group):
            raise TypeError(
                "The dataset zarr_root is not a valid Zarr archive. "
                "Did you call Dataset.load_to_memory() twice?"
            )

        # NOTE (cwognum): If the dataset fits in memory, the most performant is to use plain NumPy arrays.
        # Even if we disable chunking and compression in Zarr.
        # For more information, see https://github.com/zarr-developers/zarr-python/issues/1395
        self._zarr_data = load_zarr_group_to_memory(data)

    @abc.abstractmethod
    def get_data(
        self, row: str | int, col: str, adapters: dict[str, Adapter] | None = None
    ) -> np.ndarray | Any:
        """Since the dataset might contain pointers to external files, data retrieval is more complicated
        than just indexing the `table` attribute. This method provides an end-point for seamlessly
        accessing the underlying data.

        Args:
            row: The row index in the `Dataset.table` attribute
            col: The column index in the `Dataset.table` attribute
            adapters: The adapters to apply to the data before returning it.
                If None, will use the default adapters specified for the dataset.

        Returns:
            A numpy array with the data at the specified indices. If the column is a pointer column,
                the content of the referenced file is loaded to memory.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def upload_to_hub(
        self,
        access: AccessType = "private",
        owner: HubOwner | str | None = None,
        parent_artifact_id: str | None = None,
    ):
        """Uploads the dataset to the Polaris Hub."""
        raise NotImplementedError

    @classmethod
    @abc.abstractmethod
    def from_json(cls, path: str):
        """
        Loads a dataset from a JSON file.

        Args:
            path: The path to the JSON file to load the dataset from.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def to_json(self, destination: str | Path, if_exists: ZarrConflictResolution = "replace") -> str:
        """
        Save the dataset to a destination directory as a JSON file.

        Args:
            destination: The _directory_ to save the associated data to.
            if_exists: Action for handling existing files in the Zarr archive. Options are 'raise' to throw
                an error, 'replace' to overwrite, or 'skip' to proceed without altering the existing files.

        Returns:
            The path to the JSON file.
        """
        raise NotImplementedError

    def size(self) -> tuple[int, int]:
        return self.n_rows, self.n_columns

    def __getitem__(self, item: DatasetIndex) -> Any | np.ndarray | dict[str, np.ndarray]:
        """Allows for indexing the dataset directly"""

        # If a tuple, we assume it's the row and column index pair
        if isinstance(item, tuple):
            row, col = item
            return self.get_data(row, col)

        # Otherwise, we assume you're indexing the row
        return {col: self.get_data(item, col) for col in self.columns}

    @abc.abstractmethod
    def _repr_dict_(self) -> dict:
        """Utility function for pretty-printing to the command line and jupyter notebooks"""
        raise NotImplementedError

    def _repr_html_(self) -> str:
        """For pretty-printing in Jupyter Notebooks"""
        return dict2html(self._repr_dict_())

    def __len__(self):
        return self.n_rows

    def __repr__(self):
        return json.dumps(self._repr_dict_(), indent=2)

    def __str__(self):
        return self.__repr__()

    def _cache_zarr(self, destination: str | PathLike, if_exists: ZarrConflictResolution):
        """
        Caches the Zarr archive to a local directory.
        """
        from polaris.hub.storage import S3Store

        if not self.uses_zarr:
            return

        current_zarr_root = PurePath(self.zarr_root_path)
        destination_zarr_root = PurePath(destination) / current_zarr_root.name

        # Copy over Zarr data to the destination
        self._warn_about_remote_zarr = False

        with track_progress(description="Copying Zarr archive", total=1) as (
            progress,
            task,
        ):
            progress.log(f"[green]Copying to destination {destination_zarr_root}")
            progress.log("[yellow]For large Zarr archives, this may take a while.")
            destination_store = zarr.open(str(destination_zarr_root), "w").store
            source_store = self.zarr_root.store.store

            if isinstance(source_store, S3Store):
                source_store.copy_to_destination(destination_store, if_exists, logger.info)
            else:
                zarr.copy_store(
                    source=source_store,
                    dest=destination_store,
                    log=logger.info,
                    if_exists=if_exists,
                )
            self.zarr_root_path = str(destination_zarr_root)
            self._zarr_root = None
            self._zarr_data = None



from enum import Enum
from typing import Annotated, Any, Literal, Optional

import numpy as np
from pydantic import (
    AnyUrl,
    BaseModel,
    BeforeValidator,
    ConfigDict,
    HttpUrl,
    StringConstraints,
    TypeAdapter,
)
from pydantic.alias_generators import to_camel
from typing_extensions import Self, TypeAlias

SplitIndicesType: TypeAlias = list[int]
"""
A split is defined by a sequence of integers.
"""

SplitType: TypeAlias = tuple[SplitIndicesType, SplitIndicesType | dict[str, SplitIndicesType]]
"""
A split is a pair of which the first item is always assumed to be the train set.
The second item can either be a single test set or a dictionary with multiple, named test sets.
"""

ListOrArrayType: TypeAlias = list | np.ndarray
"""
A list of numbers or a numpy array. Predictions can be provided as either a list or a numpy array.
"""

IncomingPredictionsType: TypeAlias = ListOrArrayType | dict[str, ListOrArrayType | dict[str, ListOrArrayType]]
"""
The type of the predictions that are ingested into the Polaris BenchmarkPredictions object. Can be one
of the following:

- A single array (single-task, single test set)
- A dictionary of arrays (single-task, multiple test sets)
- A dictionary of dictionaries of arrays (multi-task, multiple test sets)
"""

PredictionsType: TypeAlias = dict[str, dict[str, np.ndarray]]
"""
The normalized format for predictions for internal use. Predictions are accepted in a generous
variety of representations and normalized into this standard format, a dictionary of dictionaries
that looks like {"test_set_name": {"target_name": np.ndarray}}.
"""

DatapointPartType = Any | tuple[Any] | dict[str, Any]
DatapointType: TypeAlias = tuple[DatapointPartType, DatapointPartType]
"""
A datapoint has:

- A single input or multiple inputs (either as dict or tuple)
- No target, a single target or a multiple targets (either as dict or tuple)
"""

SlugStringType: TypeAlias = Annotated[
    str, StringConstraints(pattern="^[a-z0-9-]+$", min_length=4, max_length=64)
]
"""
A URL-compatible string that can serve as slug on the Hub.
"""

SlugCompatibleStringType: TypeAlias = Annotated[
    str, StringConstraints(pattern="^[A-Za-z0-9_-]+$", min_length=4, max_length=64)
]
"""
A URL-compatible string that can be turned into a slug by the Hub.

Can only use alpha-numeric characters, underscores and dashes.
The string must be at least 4 and at most 64 characters long.
"""

Md5StringType: TypeAlias = Annotated[str, StringConstraints(pattern=r"^[a-f0-9]{32}$")]
"""
A string that represents an MD5 hash.
"""

HubUser: TypeAlias = SlugCompatibleStringType
"""
A user on the Polaris Hub is identified by a username,
which is a [`SlugCompatibleStringType`][polaris.utils.types.SlugCompatibleStringType].
"""

HttpUrlAdapter = TypeAdapter(HttpUrl)
HttpUrlString: TypeAlias = Annotated[str, BeforeValidator(lambda v: HttpUrlAdapter.validate_python(v) and v)]
"""
A validated HTTP URL that will be turned into a string.
This is useful for interactions with httpx and authlib, who have their own URL types.
"""

AnyUrlAdapter = TypeAdapter(AnyUrl)
AnyUrlString: TypeAlias = Annotated[str, BeforeValidator(lambda v: AnyUrlAdapter.validate_python(v) and v)]
"""
A validated generic URL that will be turned into a string.
This is useful for interactions with other libraries that expect a string.
"""

DirectionType: TypeAlias = float | Literal["min", "max"]
"""
The direction of any variable to be sorted.
This can be used to sort the metric score, indicate the optmization direction of endpoint.
"""

AccessType: TypeAlias = Literal["public", "private"]
"""
Type to specify access to a dataset, benchmark or result in the Hub.
"""

TimeoutTypes = tuple[int, int] | Literal["timeout", "never"]
"""
Timeout types for specifying maximum wait times.
"""

IOMode: TypeAlias = Literal["r", "r+", "a", "w", "w-"]
"""
Type to specify the mode for input/output operations (I/O) when interacting with a file or resource.
"""

SupportedLicenseType: TypeAlias = Literal[
    "CC-BY-4.0", "CC-BY-SA-4.0", "CC-BY-NC-4.0", "CC-BY-NC-SA-4.0", "CC0-1.0", "MIT"
]
"""
Supported license types for dataset uploads to Polaris Hub
"""

ZarrConflictResolution: TypeAlias = Literal["raise", "replace", "skip"]
"""
Type to specify which action to take when encountering existing files within a Zarr archive.
"""

ChecksumStrategy: TypeAlias = Literal["verify", "verify_unless_zarr", "ignore"]
"""
Type to specify which action to take to verify the data integrity of an artifact through a checksum.
"""

ArtifactUrn: TypeAlias = Annotated[str, StringConstraints(pattern=r"^urn:polaris:\w+:\w+:\w+$")]
"""
A Uniform Resource Name (URN) for an artifact on the Polaris Hub.
"""

RowIndex: TypeAlias = int | str
ColumnIndex: TypeAlias = str
DatasetIndex: TypeAlias = RowIndex | tuple[RowIndex, ColumnIndex]
"""
To index a dataset using square brackets, we have a few options:

- A single row, e.g. dataset[0]
- Specify a specific value, e.g. dataset[0, "col1"]

There are more exciting options we could implement, such as slicing,
but this gets complex.
"""


PredictionKwargs: TypeAlias = Literal["y_pred", "y_prob", "y_score"]
"""
The type of predictions expected by the metric interface.
"""

ColumnName: TypeAlias = str
"""A column name in a dataset."""


class HubOwner(BaseModel):
    """An owner of an artifact on the Polaris Hub

    The slug is most important as it is the user-facing part of this data model.
    The externalId and type are added to be consistent with the model returned by the Polaris Hub .
    """

    slug: SlugStringType
    external_id: Optional[str] = None
    type: Optional[Literal["user", "organization"]] = None

    # Pydantic config
    model_config = ConfigDict(alias_generator=to_camel, populate_by_name=True)

    def __str__(self):
        return self.slug

    @staticmethod
    def normalize(owner: str | Self) -> Self:
        """
        Normalize a string or `HubOwner` instance to a `HubOwner` instance.
        """
        return owner if isinstance(owner, HubOwner) else HubOwner(slug=owner)


class TargetType(Enum):
    """The high-level classification of different targets."""

    REGRESSION = "regression"
    CLASSIFICATION = "classification"
    DOCKING = "docking"


class TaskType(Enum):
    """The high-level classification of different tasks."""

    MULTI_TASK = "multi_task"
    SINGLE_TASK = "single_task"



import logging
import webbrowser
from typing import Literal, Optional, TypeAlias

from authlib.common.security import generate_token
from authlib.integrations.base_client import OAuthError
from authlib.integrations.httpx_client import OAuth2Client
from authlib.oauth2 import OAuth2Error, TokenAuth
from authlib.oauth2.rfc6749 import OAuth2Token

from polaris.hub.oauth import ExternalCachedTokenAuth
from polaris.hub.settings import PolarisHubSettings
from polaris.utils.errors import PolarisHubError, PolarisUnauthorizedError

logger = logging.getLogger(__name__)

Scope: TypeAlias = Literal["read", "write"]


class ExternalAuthClient(OAuth2Client):
    """
    This authentication client is used to obtain OAuth 2 tokens from Polaris's external OAuth2 server.
    These can in turn be used to obtain Polaris Hub tokens.

    Note: Internal use
        This class is intended for internal use by the `PolarisHubClient` class, and you should not have to
        interact with it directly.
    """

    def __init__(
        self,
        settings: PolarisHubSettings,
        cache_auth_token: bool = True,
        **kwargs: dict,
    ):
        """
        Args:
            settings: A `PolarisHubSettings` instance.
            cache_auth_token: Whether to cache the auth token to a file.
            **kwargs: Additional keyword arguments passed to the authlib `OAuth2Client` constructor.
        """
        self._user_info = None

        # We cache the auth token by default, but allow the user to disable this.
        self.token_auth_class = ExternalCachedTokenAuth if cache_auth_token else TokenAuth

        self.settings = settings

        self.code_verifier = generate_token(48)

        super().__init__(
            # OAuth2Client
            client_id=self.settings.client_id,
            redirect_uri=self.settings.callback_url,
            scope=self.settings.scopes,
            token_endpoint=self.settings.token_fetch_url,
            code_challenge_method="S256",
            # httpx.Client
            timeout=self.settings.default_timeout,
            cert=self.settings.ca_bundle,
            # Extra
            **kwargs,
        )

    def create_authorization_url(self, **kwargs) -> tuple[str, Optional[str]]:
        """Light wrapper to automatically pass in the right URL."""
        return super().create_authorization_url(
            url=self.settings.authorize_url, code_verifier=self.code_verifier, **kwargs
        )

    def fetch_token(self, **kwargs) -> dict:
        """Light wrapper to automatically pass in the right URL"""
        try:
            return super().fetch_token(
                url=self.settings.token_fetch_url, code_verifier=self.code_verifier, **kwargs
            )
        except OAuth2Error as error:
            raise PolarisHubError(
                message=f"Could not obtain a token from the external OAuth2 server. Error was: {error.error} - {error.description}"
            ) from error

    def ensure_active_token(self, token: OAuth2Token | None = None) -> bool:
        try:
            # This won't be needed with if we set a lower bound for authlib: >=1.3.2
            # See https://github.com/lepture/authlib/pull/625
            # As of now, this latest version is not available on Conda though.
            token = token or self.token
            return super().ensure_active_token(token) if token else False
        except OAuthError:
            # The refresh attempt can fail with this error
            return False

    @property
    def user_info(self) -> dict:
        """
        Get information about the currently logged-in user through the OAuth2 User Info flow."""

        # NOTE (cwognum): We override the default `auth` and `headers` argument, since
        #  the defaults trigger a 530 error (Cloudflare) due to the header ordering.
        #  Because of this, we also have to copy some code from the base `request` method to
        #  make auto-refresh a token if needed. For more info, see: https://stackoverflow.com/a/62687390

        try:
            if self.token is None or not self.ensure_active_token(self.token):
                raise PolarisUnauthorizedError
        except OAuthError:
            raise PolarisUnauthorizedError

        if self._user_info is None:
            user_info = self.get(
                self.settings.user_info_url,
                auth=None,  # type: ignore
                headers={"authorization": f"Bearer {self.token['access_token']}"},
            )
            user_info.raise_for_status()
            self._user_info = user_info.json()

        return self._user_info

    def interactive_login(self, overwrite: bool = False, auto_open_browser: bool = True):
        """
        Login to the Polaris Hub using an interactive flow, through a Web browser.

        Warning: Headless authentication
            It is currently not possible to log in to the Polaris Hub without a browser.
            See [this GitHub issue](https://github.com/polaris-hub/polaris/issues/30) for more info.

        Args:
            overwrite: Whether to overwrite the current token if the user is already logged in.
            auto_open_browser: Whether to automatically open the browser to visit the authorization URL.
        """

        # Check if the user is already logged in
        token_is_valid = self.token is not None and self.ensure_active_token(self.token)
        if token_is_valid and not overwrite:
            try:
                info = self.user_info
                logger.info(
                    f"You are already logged in to the Polaris Hub as {info['email']}. Set `overwrite=True` to force re-authentication."
                )
                return
            except PolarisUnauthorizedError:
                pass

        # Step 1: Redirect user to the authorization URL
        authorization_url, _ = self.create_authorization_url()

        if auto_open_browser:
            logger.info(f"Your browser has been opened to visit:\n{authorization_url}\n")
            webbrowser.open_new_tab(authorization_url)
        else:
            logger.info(f"Please visit the following URL:\n{authorization_url}\n")

        # Step 2: After user grants permission, we'll get the authorization code through the callback URL
        authorization_code = input("Please enter the authorization token: ")

        # Step 3: Exchange authorization code for an access token
        self.fetch_token(code=authorization_code, grant_type="authorization_code")

        logger.info(f"Successfully authenticated to the Polaris Hub as `{self.user_info['email']}`!")



from collections import defaultdict

import numpy as np
from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    TypeAdapter,
    field_serializer,
    field_validator,
    model_validator,
)
from typing_extensions import Self

from polaris.evaluate import ResultsMetadataV1
from polaris.utils.misc import convert_lists_to_arrays
from polaris.utils.types import (
    HttpUrlString,
    HubOwner,
    IncomingPredictionsType,
    PredictionsType,
    SlugCompatibleStringType,
)


class BenchmarkPredictions(BaseModel):
    """
    Base model to represent predictions in the Polaris code base.

    Guided by [Postel's Law](https://en.wikipedia.org/wiki/Robustness_principle),
    this class normalizes different formats to a single, internal representation.

    Attributes:
        predictions: The predictions for the benchmark.
        target_labels: The target columns for the associated benchmark.
        test_set_labels: The names of the test sets for the associated benchmark.
        test_set_sizes: The number of rows in each test set for the associated benchmark.
    """

    predictions: PredictionsType
    target_labels: list[str]
    test_set_labels: list[str]
    test_set_sizes: dict[str, int]

    model_config = ConfigDict(arbitrary_types_allowed=True)

    @field_serializer("predictions")
    def _serialize_predictions(self, predictions: PredictionsType):
        """
        Recursively converts all numpy values in the predictions dictionary to lists
        so they can be serialized.
        """

        def convert_to_list(v):
            if isinstance(v, np.ndarray):
                return v.tolist()
            elif isinstance(v, dict):
                return {k: convert_to_list(v) for k, v in v.items()}

        return convert_to_list(predictions)

    @field_validator("target_labels", "test_set_labels")
    @classmethod
    def _validate_labels(cls, v: list[str]) -> list[str]:
        if len(set(v)) != len(v):
            raise ValueError("The predictions contain duplicate columns")
        return v

    @model_validator(mode="before")
    @classmethod
    def _validate_predictions(cls, data: dict) -> dict:
        """Normalizes the predictions format to a standard representation we use internally"""

        # This model validator runs before any Pydantic internal validation.
        # This way we can normalize the incoming data to a standard representation.
        # However, this implies that the fields can theoretically be any type.

        # Ensure the type of the incoming predictions is correct
        validator = TypeAdapter(IncomingPredictionsType, config={"arbitrary_types_allowed": True})
        predictions = validator.validate_python(data.get("predictions"))

        # Ensure the type of the target_labels and test_set_labels is correct
        validator = TypeAdapter(list[str])
        target_labels = validator.validate_python(data.get("target_labels"))
        test_set_labels = validator.validate_python(data.get("test_set_labels"))

        validator = TypeAdapter(dict[str, int])
        test_set_sizes = validator.validate_python(data.get("test_set_sizes"))

        # Normalize the predictions to a standard representation
        predictions = convert_lists_to_arrays(predictions)
        predictions = cls._normalize_predictions(predictions, target_labels, test_set_labels)

        # Update class data with the normalized fields. Use of the `update()` method
        # is required to prevent overwriting class data when this class is inherited.
        data.update(
            {
                "predictions": predictions,
                "target_labels": target_labels,
                "test_set_labels": test_set_labels,
                "test_set_sizes": test_set_sizes,
            }
        )

        return data

    @model_validator(mode="after")
    def check_test_set_size(self) -> Self:
        """Verify that the size of all predictions"""
        for test_set_label, test_set in self.predictions.items():
            for target in test_set.values():
                if test_set_label not in self.test_set_sizes:
                    raise ValueError(f"Expected size for test set '{test_set_label}' is not defined")

                if len(target) != self.test_set_sizes[test_set_label]:
                    raise ValueError(
                        f"Predictions size mismatch: The predictions for test set '{test_set_label}' "
                        f"should have a size of {self.test_set_sizes[test_set_label]}, but have a size of {len(target)}."
                    )
        return self

    @classmethod
    def _normalize_predictions(
        cls, predictions: IncomingPredictionsType, target_labels: list[str], test_set_labels: list[str]
    ) -> PredictionsType:
        """
        Normalizes  the predictions to a standard representation we use internally.
        This standard representation is a nested, two-level dictionary:
        `{test_set_name: {target_column: np.ndarray}}`
        """
        # (1) If the predictions are already fully specified, no need to do anything
        if cls._is_fully_specified(predictions, target_labels, test_set_labels):
            return predictions

        # If not fully specified, we distinguish 4 cases based on the type of benchmark.
        is_single_task = len(target_labels) == 1
        is_single_test = len(test_set_labels) == 1

        # (2) Single-task, single test set: We expect a numpy array as input.
        if is_single_task and is_single_test:
            if isinstance(predictions, dict):
                raise ValueError(
                    "The predictions for single-task, single test set benchmarks should be a numpy array."
                )

            predictions = {test_set_labels[0]: {target_labels[0]: predictions}}

        # (3) Single-task, multiple test sets: We expect a dictionary with the test set labels as keys.
        elif is_single_task and not is_single_test:
            if not isinstance(predictions, dict) or set(predictions.keys()) != set(test_set_labels):
                raise ValueError(
                    "The predictions for single-task, multiple test sets benchmarks "
                    "should be a dictionary with the test set labels as keys."
                )
            predictions = {k: {target_labels[0]: v} for k, v in predictions.items()}

        # (4) Multi-task, single test set: We expect a dictionary with the target labels as keys.
        elif not is_single_task and is_single_test:
            if not isinstance(predictions, dict) or set(predictions.keys()) != set(target_labels):
                raise ValueError(
                    "The predictions for multi-task, single test set benchmarks "
                    "should be a dictionary with the target labels as keys."
                )
            predictions = {test_set_labels[0]: predictions}

        # (5) Multi-task, multi-test sets: The predictions should be fully-specified
        else:
            raise ValueError(
                "The predictions for multi-task, multi-test sets benchmarks should be fully-specified "
                "as a nested, two-level dictionary: { test_set_name: { target_column: np.ndarray } }"
            )

        return predictions

    @classmethod
    def _is_fully_specified(
        cls, predictions: IncomingPredictionsType, target_labels: list[str], test_set_labels: list[str]
    ) -> bool:
        """
        Check if the predictions are fully specified for the target columns and test set names.
        """
        # Not a dictionary
        if not isinstance(predictions, dict):
            return False

        # Outer-level of the dictionary should correspond to the test set names
        if set(predictions.keys()) != set(test_set_labels):
            return False

        # Inner-level of the dictionary should correspond to the target columns
        for test_set_predictions in predictions.values():
            if not isinstance(test_set_predictions, dict):
                return False
            if set(test_set_predictions.keys()) != set(target_labels):
                return False

        return True

    def get_subset(
        self, test_set_subset: list[str] | None = None, target_subset: list[str] | None = None
    ) -> "BenchmarkPredictions":
        """Return a subset of the original predictions"""

        if test_set_subset is None and target_subset is None:
            return self

        if test_set_subset is None:
            test_set_subset = self.test_set_labels
        if target_subset is None:
            target_subset = self.target_labels

        predictions = defaultdict(dict)
        for test_set_name in self.predictions.keys():
            if test_set_name not in test_set_subset:
                continue

            for target_name, target in self.predictions[test_set_name].items():
                if target_name not in target_subset:
                    continue

                predictions[test_set_name][target_name] = target

        test_set_sizes = {k: v for k, v in self.test_set_sizes.items() if k in predictions.keys()}
        return BenchmarkPredictions(
            predictions=predictions,
            target_labels=target_subset,
            test_set_labels=test_set_subset,
            test_set_sizes=test_set_sizes,
        )

    def get_size(
        self, test_set_subset: list[str] | None = None, target_subset: list[str] | None = None
    ) -> int:
        """Return the total number of predictions, allowing for filtering by test set and target"""
        if test_set_subset is None and target_subset is None:
            return sum(self.test_set_sizes.values()) * len(self.target_labels)
        return len(self.get_subset(test_set_subset, target_subset))

    def flatten(self) -> np.ndarray:
        """Return the predictions as a single, flat numpy array"""
        if len(self.test_set_labels) != 1 and len(self.target_labels) != 1:
            raise ValueError(
                "Can only flatten predictions for benchmarks with a single test set and a single target"
            )
        return self.predictions[self.test_set_labels[0]][self.target_labels[0]]

    def __len__(self) -> int:
        """Return the total number of predictions"""
        return self.get_size()


class CompetitionPredictions(BenchmarkPredictions, ResultsMetadataV1):
    """
    Predictions for competition benchmarks.

    This object is to be used as input to
    [`PolarisHubClient.submit_competition_predictions`][polaris.hub.client.PolarisHubClient.submit_competition_predictions].
    It is used to ensure that the structure of the predictions are compatible with evaluation methods on the Polaris Hub.
    In addition to the predictions, it contains metadata that describes a predictions object.

    Attributes:
        name: A slug-compatible name for the artifact. It is redeclared here to be required.
        owner: A slug-compatible name for the owner of the artifact. It is redeclared here to be required.
        report_url: A URL to a report/paper/write-up which describes the methods used to generate the predictions.
    """

    _artifact_type = "competition-prediction"

    name: SlugCompatibleStringType
    owner: HubOwner
    paper_url: HttpUrlString = Field(alias="report_url", serialization_alias="reportUrl")

    def __repr__(self):
        return self.model_dump_json(by_alias=True, indent=2)

    def __str__(self):
        return self.__repr__()



import re
from base64 import b64encode
from concurrent.futures import ThreadPoolExecutor, as_completed
from contextlib import contextmanager
from functools import lru_cache
from hashlib import md5
from itertools import islice
from pathlib import PurePath
from typing import Any, Callable, Generator, Literal, Mapping, Sequence, TypeAlias

import boto3
from authlib.integrations.httpx_client import OAuth2Client, OAuthError
from authlib.oauth2 import OAuth2Error
from authlib.oauth2.rfc6749 import OAuth2Token
from botocore.exceptions import BotoCoreError, ClientError
from typing_extensions import Self
from zarr import CopyError
from zarr.context import Context
from zarr.storage import Store
from zarr.util import buffer_size

from polaris.hub.oauth import BenchmarkV2Paths, DatasetV1Paths, DatasetV2Paths, HubStorageOAuth2Token
from polaris.utils.context import track_progress
from polaris.utils.errors import PolarisHubError
from polaris.utils.types import ArtifactUrn, ZarrConflictResolution

Scope: TypeAlias = Literal["read", "write"]


class S3StoreException(Exception):
    """
    Base exception for S3Store.
    """


class S3StoreCredentialsExpiredException(S3StoreException):
    """
    Exception raised when the S3 credentials have expired.
    """


@contextmanager
def handle_s3_errors():
    """
    Standardize error handling for S3 operations.
    """
    try:
        yield
    except ClientError as e:
        error_code = e.response["Error"]["Code"]
        if error_code == "ExpiredToken":
            raise S3StoreCredentialsExpiredException(f"Error in S3Store: Credentials expired: {e}") from e
        else:
            raise S3StoreException(f"Error in S3Store: {e}") from e
    except BotoCoreError as e:
        raise S3StoreException(f"Error in S3Store: {e}") from e


class S3Store(Store):
    """
     A Zarr store implementation using a S3 bucket as the backend storage.

    It supports multipart uploads for large objects and handles S3-specific exceptions.
    """

    _erasable = False
    _batch_size = 200

    def __init__(
        self,
        path: str | PurePath,
        access_key: str,
        secret_key: str,
        token: str,
        endpoint_url: str,
        part_size: int = 10 * 1024 * 1024,  # 10MB
        content_type: str = "application/octet-stream",
    ) -> None:
        bucket_name, *prefix = PurePath(path).parts

        self.s3_client = boto3.client(
            "s3",
            aws_access_key_id=access_key,
            aws_secret_access_key=secret_key,
            aws_session_token=token,
            endpoint_url=endpoint_url,
            region_name="auto",
        )
        self.bucket_name = bucket_name
        self.prefix = "/".join(prefix)
        self.part_size = part_size
        self.content_type = content_type

    def _full_key(self, key: str) -> str:
        """
        Converts a relative key to the full bucket key.
        """
        return f"{self.prefix}/{key}"

    def _multipart_upload(self, key: str, value: bytes) -> None:
        """
        For large files, use multipart to split the work.
        """
        full_key = self._full_key(key)
        md5_hash = md5(value)
        with handle_s3_errors():
            upload = self.s3_client.create_multipart_upload(
                Bucket=self.bucket_name,
                Key=full_key,
                ContentType=self.content_type,
                Metadata={
                    "md5sum": md5_hash.hexdigest(),
                },
            )
            upload_id = upload["UploadId"]

            parts = []
            for i in range(0, len(value), self.part_size):
                part_number = i // self.part_size + 1
                part = value[i : i + self.part_size]
                response = self.s3_client.upload_part(
                    Bucket=self.bucket_name,
                    Key=full_key,
                    PartNumber=part_number,
                    UploadId=upload_id,
                    Body=part,
                    ContentMD5=b64encode(md5(part).digest()).decode(),
                )
                parts.append({"ETag": response["ETag"], "PartNumber": part_number})

            self.s3_client.complete_multipart_upload(
                Bucket=self.bucket_name, Key=full_key, UploadId=upload_id, MultipartUpload={"Parts": parts}
            )

    @lru_cache()
    def _get_object_body(self, full_key: str) -> bytes:
        """
        Basic caching for the object body, to avoid multiple reads on remote bucket.
        """
        response = self.s3_client.get_object(Bucket=self.bucket_name, Key=full_key)
        return response["Body"].read()

    ## Custom methods

    def copy_to_destination(
        self, destination: Store, if_exists: ZarrConflictResolution = "replace", log: Callable = lambda: None
    ) -> tuple[int, int, int]:
        """
        Copy the content of this store to the destination store.

        We leverage the internal knowledge of this store to make the operation more efficient than `zarr.copy_store`:
            - Parallel, concurrent `getitems` operations using a thread pool

        Zarr V3 supports partial writes, that would allow us to stream the response back into the store. Unfortunately,
        it's not supported right now by the library version we use, so we'll have to read the whole response into memory
        and write it back to the destination.

        Args:
            destination: destination store
            if_exists: behavior if the destination key already exists
            log: optional logging function
        """

        destination_store_version = getattr(destination, "_store_version", 2)
        if destination_store_version != self._store_version:
            raise ValueError("zarr stores must share the same protocol version")

        total_copied = total_skipped = total_bytes_copied = 0

        number_source_keys = len(self)

        with track_progress(description="Copying Zarr keys", total=number_source_keys) as (
            progress_keys,
            task_keys,
        ):
            batch_iter = iter(self)
            while batch := tuple(islice(batch_iter, self._batch_size)):
                to_put = (
                    batch if if_exists == "replace" else filter(lambda key: key not in destination, batch)
                )
                skipped = len(batch) - len(to_put)

                if skipped > 0:
                    if if_exists == "raise":
                        raise CopyError(f"keys {to_put} exist in destination")
                    else:
                        progress_keys.log(f"Skipped {skipped} keys that already exists")

                items = self.getitems(to_put, contexts={})
                for key, content in items.items():
                    destination[key] = content

                    size_copied = buffer_size(content)
                    total_bytes_copied += size_copied
                    total_copied += 1

                total_skipped += skipped
                progress_keys.update(task_keys, advance=len(batch), refresh=True)

        return total_copied, total_skipped, total_bytes_copied

    def copy_from_source(
        self, source: Store, if_exists: ZarrConflictResolution = "replace", log: Callable = lambda: None
    ) -> tuple[int, int, int]:
        """
        Copy the content of the source store to this store.

        We leverage the internal knowledge of this store to make the operation more efficient than `zarr.copy_store`:
            - Conditional `put_object` operation for the "skip" conflict resolution
            - Parallel, concurrent `put_object` operations using a thread pool

        Args:
            source: source store
            if_exists: behavior if the destination key already exists
            log: optional logging function
        """

        def copy_key(key: str, source: Store, if_exists: ZarrConflictResolution) -> tuple[int, int, int]:
            """
            Sub-function, intended to be dispatched through a thread pool executor.
            It will execute the copy operation for a single key from the source.
            """
            copied = skipped = bytes_copied = 0

            with handle_s3_errors():
                data = source[key]
                if isinstance(data, memoryview):
                    data = data.tobytes()

                md5_hash = md5(data)
                try:
                    # The default is to replace the existing key.
                    # Setting this header will raise an error if the key already exists.
                    extra = {"IfNoneMatch": "*"} if if_exists == "skip" else {}
                    self.s3_client.put_object(
                        Bucket=self.bucket_name,
                        Key=self._full_key(key),
                        Body=data,
                        ContentType=self.content_type,
                        ContentMD5=b64encode(md5_hash.digest()).decode(),
                        Metadata={
                            "md5sum": md5_hash.hexdigest(),
                        },
                        **extra,
                    )
                    bytes_copied = buffer_size(data)
                    copied = 1
                except ClientError as error:
                    error_code = error.response["Error"]["Code"]
                    # Raised when the key already exists and IfNoneMatch is set
                    if error_code == "PreconditionFailed":
                        match if_exists:
                            case "skip":
                                skipped = 1
                            case "raise":
                                raise CopyError(f"key {key!r} exists in destination")
                    else:
                        raise
            return copied, skipped, bytes_copied

        source_store_version = getattr(source, "_store_version", 2)
        if source_store_version != self._store_version:
            raise ValueError("zarr stores must share the same protocol version")

        # The executor will allocate min(nbr_CPU + 4, 32) threads
        with ThreadPoolExecutor() as executor:
            total_copied = total_skipped = total_bytes_copied = 0

            number_source_keys = len(source)

            with track_progress(description="Copying Zarr keys", total=number_source_keys) as (
                progress_keys,
                task_keys,
            ):
                # Batch the keys, otherwise we end up with too many files open at the same time
                batch_iter = iter(source.keys())
                while batch := tuple(islice(batch_iter, self._batch_size)):
                    # Create a future for each key to copy
                    future_to_key = [
                        executor.submit(copy_key, source_key, source, if_exists) for source_key in batch
                    ]

                    # As each future completes, collect the results
                    for future in as_completed(future_to_key):
                        result_copied, result_skipped, result_bytes_copied = future.result()
                        total_copied += result_copied
                        progress_keys.update(task_keys, advance=result_copied, refresh=True)

                        total_skipped += result_skipped
                        if result_skipped > 0:
                            progress_keys.log(f"Skipped {result_skipped} keys that already exists")

                        total_bytes_copied += result_bytes_copied

            return total_copied, total_skipped, total_bytes_copied

    ## Optional Zarr store methods

    def listdir(self, path: str = "") -> Generator[str, None, None]:
        """
        For a given path, list all the "subdirectories" and "files" for that path.
        The returned paths are relative to the input path.

        Uses pagination and return a generator to handle very large number of keys.
        Note: This might not help with some Zarr operations that materialize the whole sequence.
        """
        prefix = self._full_key(path)

        # Ensure a trailing slash to avoid the path looking like one specific key
        if prefix and not prefix.endswith("/"):
            prefix += "/"

        with handle_s3_errors():
            # `list_objects_v2` returns a max of 1000 keys per request, so paginate requests
            paginator = self.s3_client.get_paginator("list_objects_v2")
            page_iterator = paginator.paginate(Bucket=self.bucket_name, Prefix=prefix, Delimiter="/")

            for page in page_iterator:
                # Contents are "files"
                for obj in page.get("Contents", []):
                    key = obj["Key"][len(prefix) :]
                    if key:
                        yield key.split("/")[0]

                # CommonPrefixes are "subdirectories"
                for common_prefix in page.get("CommonPrefixes", []):
                    yield common_prefix["Prefix"][len(prefix) :].strip("/")

    def getitems(self, keys: Sequence[str], *, contexts: Mapping[str, Context]) -> dict[str, Any]:
        """
        More efficient implementation of getitems using concurrent fetching through multiple threads.

        The default implementation uses __contains__ to check existence before fetching the value,
        which doubles the number of requests.
        """

        def fetch_key(key: str) -> tuple[str, Any]:
            """
            Wrapper that looks up a key's value in the store
            """
            try:
                return key, self[key]
            except KeyError:
                return key, None

        results = {}
        with ThreadPoolExecutor() as executor:
            # Create a future for each key to look up
            future_to_key = {executor.submit(fetch_key, key): key for key in keys}

            # As each future completes, collect the results, if any
            for future in as_completed(future_to_key):
                key, value = future.result()
                if value is not None:
                    results[key] = value

        return results

    def getsize(self, key: str) -> int:
        """
        Return the size (in bytes) of the object at the given key.
        """
        with handle_s3_errors():
            response = self.s3_client.head_object(Bucket=self.bucket_name, Key=self._full_key(key))
            return response["ContentLength"]

    ## MutableMapping implementation, expected by Zarr store

    def __getitem__(self, key: str) -> bytes:
        """
        Retrieves the value for the given key from the store.

        Makes no provision to handle overly large values returned.
        """
        with handle_s3_errors():
            try:
                full_key = self._full_key(key)
                return self._get_object_body(full_key=full_key)
            except self.s3_client.exceptions.NoSuchKey:
                raise KeyError(key)

    def __setitem__(self, key: str, value: bytes | bytearray | memoryview) -> None:
        """
        Persists the given value in the store.

        Based on value size, will use multipart upload for large files,
        or a single put_object call.
        """
        if isinstance(value, memoryview):
            value = value.tobytes()

        if len(value) > self.part_size:
            self._multipart_upload(key, value)
        else:
            with handle_s3_errors():
                md5_hash = md5(value)
                self.s3_client.put_object(
                    Bucket=self.bucket_name,
                    Key=self._full_key(key),
                    Body=value,
                    ContentType=self.content_type,
                    ContentMD5=b64encode(md5_hash.digest()).decode(),
                    Metadata={
                        "md5sum": md5_hash.hexdigest(),
                    },
                )

    def __delitem__(self, key: str) -> None:
        """
        Removing a key from the store is not supported.
        """
        raise NotImplementedError(f'{type(self)} is not erasable, cannot call "del store[key]"')

    def __contains__(self, key: str) -> bool:
        """
        Checks the existence of a key in the store.

        If the intent is to download the value after this check, it is more efficient to
        attempt tp retrieve it and handle the KeyError from a non-existent key.
        """
        with handle_s3_errors():
            try:
                self.s3_client.head_object(Bucket=self.bucket_name, Key=self._full_key(key))
                return True
            except self.s3_client.exceptions.NoSuchKey:
                return False
            except ClientError as e:
                if e.response["Error"]["Code"] == "404":
                    return False
                raise e

    def __iter__(self) -> Generator[str, None, None]:
        """
        Iterate through all the keys in the store.
        """
        with handle_s3_errors():
            # `list_objects_v2` returns a max of 1000 keys per request, so paginate requests
            paginator = self.s3_client.get_paginator("list_objects_v2")
            page_iterator = paginator.paginate(Bucket=self.bucket_name, Prefix=self.prefix)

            for page in page_iterator:
                for obj in page.get("Contents", []):
                    yield obj["Key"][len(self.prefix) + 1 :]

    def __len__(self) -> int:
        """
        Number of keys in the store.
        """
        with handle_s3_errors():
            # `list_objects_v2` returns a max of 1000 keys per request, so paginate requests
            paginator = self.s3_client.get_paginator("list_objects_v2")
            page_iterator = paginator.paginate(Bucket=self.bucket_name, Prefix=self.prefix)

            return sum((page["KeyCount"] for page in page_iterator))

    def __hash__(self):
        """
        Custom hash function, to enable lru_cache decorator on methods
        """
        return hash((self.bucket_name, self.prefix, self.s3_client))


class StorageTokenAuth:
    token: HubStorageOAuth2Token | None

    def __init__(self, token: dict[str, Any] | None, *args) -> None:
        self.token = None
        if token:
            self.set_token(token)

    def set_token(self, token: dict[str, Any] | HubStorageOAuth2Token) -> None:
        self.token = HubStorageOAuth2Token(**token) if isinstance(token, dict) else token


class StorageSession(OAuth2Client):
    """
    A context manager for managing a storage session, with token exchange and token refresh capabilities.
    Each session is associated with a specific scope and resource.
    """

    polaris_protocol = "polarisfs"

    token_auth_class = StorageTokenAuth

    def __init__(self, hub_client, scope: Scope, resource: ArtifactUrn):
        self.hub_client = hub_client
        self.resource = resource

        super().__init__(
            # OAuth2Client
            token_endpoint=hub_client.settings.hub_token_url,
            token_endpoint_auth_method="none",
            grant_type="urn:ietf:params:oauth:grant-type:token-exchange",
            scope=scope,
            # httpx.Client
            cert=hub_client.settings.ca_bundle,
        )

    def __enter__(self) -> Self:
        self.ensure_active_token()
        return self

    def _prepare_token_endpoint_body(self, body, grant_type, **kwargs) -> str:
        """
        Override to support required fields for the token exchange grant type.
        See https://datatracker.ietf.org/doc/html/rfc8693#name-request
        """
        if grant_type == "urn:ietf:params:oauth:grant-type:token-exchange":
            kwargs.update(
                {
                    "subject_token": self.hub_client.token.get("access_token"),
                    "subject_token_type": "urn:ietf:params:oauth:token-type:jwt",
                    "requested_token_type": "urn:ietf:params:oauth:token-type:jwt",
                    "resource": self.resource,
                }
            )
        return super()._prepare_token_endpoint_body(body, grant_type, **kwargs)

    def fetch_token(self, **kwargs) -> dict[str, Any]:
        """
        Error handling for token fetching.
        """
        try:
            return super().fetch_token(**kwargs)
        except (OAuthError, OAuth2Error) as error:
            raise PolarisHubError(
                message=f"Could not obtain a token to access the storage backend. Error was: {error.error} - {error.description}"
            ) from error

    def ensure_active_token(self, token: OAuth2Token | None = None) -> bool:
        """
        Override the active check to trigger a re-fetch of the token if it is not active.
        """
        if token is None:
            # This won't be needed with if we set a lower bound for authlib: >=1.3.2
            # See https://github.com/lepture/authlib/pull/625
            # As of now, this latest version is not available on Conda though.
            token = self.token

        if token and super().ensure_active_token(token):
            return True

        # Check if external token is still valid
        if not self.hub_client.ensure_active_token():
            return False

        # If so, use it to get a new Hub token
        self.token = self.fetch_token()
        return True

    @property
    def paths(self) -> DatasetV1Paths | DatasetV2Paths | BenchmarkV2Paths:
        return self.token.extra_data.paths

    def _relative_path(self, path: str) -> PurePath:
        return PurePath(re.sub(r"^\w+://", "", path))

    def set_file(self, path: str, value: bytes | bytearray):
        """
        Set a value at the given path.
        """
        if path not in self.paths.files:
            raise NotImplementedError(f"{type(self.paths).__name__} only supports files {self.paths.files}.")

        relative_path = self._relative_path(getattr(self.paths, path))

        match relative_path.suffix:
            case ".parquet":
                content_type = "application/vnd.apache.parquet"
            case ".roaring":
                content_type = "application/vnd.roaringbitmap"
            case _:
                content_type = "application/octet-stream"

        storage_data = self.token.extra_data
        store = S3Store(
            path=relative_path.parent,
            access_key=storage_data.key,
            secret_key=storage_data.secret,
            token=f"jwt/{self.token.access_token}",
            endpoint_url=storage_data.endpoint,
            content_type=content_type,
        )
        store[relative_path.name] = value

    def get_file(self, path: str) -> bytes | bytearray:
        """
        Get the value at the given path.
        """
        if path not in self.paths.files:
            raise NotImplementedError(
                f"{type(self.paths).__name__} only supports these files: {self.paths.files}."
            )

        relative_path = self._relative_path(getattr(self.paths, path))

        storage_data = self.token.extra_data
        store = S3Store(
            path=relative_path.parent,
            access_key=storage_data.key,
            secret_key=storage_data.secret,
            token=f"jwt/{self.token.access_token}",
            endpoint_url=storage_data.endpoint,
        )
        return store[relative_path.name]

    def store(self, path: str) -> S3Store:
        if path not in self.paths.stores:
            raise NotImplementedError(
                f"{type(self.paths).__name__} only supports these stores: {self.paths.stores}."
            )

        relative_path = self._relative_path(getattr(self.paths, path))

        storage_data = self.token.extra_data
        return S3Store(
            path=relative_path,
            access_key=storage_data.key,
            secret_key=storage_data.secret,
            token=f"jwt/{self.token.access_token}",
            endpoint_url=storage_data.endpoint,
        )



import pytest
from pydantic import ValidationError

from polaris import load_benchmark
from polaris.benchmark import (
    MultiTaskBenchmarkSpecification,
    SingleTaskBenchmarkSpecification,
)
from polaris.evaluate import Metric


@pytest.mark.parametrize("is_single_task", [True, False])
def test_split_verification(is_single_task, test_single_task_benchmark, test_multi_task_benchmark):
    """Verifies that the split validation works as expected."""

    obj = test_single_task_benchmark if is_single_task else test_multi_task_benchmark
    cls = SingleTaskBenchmarkSpecification if is_single_task else MultiTaskBenchmarkSpecification

    # By using the fixture as a default, we know it doesn't always fail
    default_kwargs = {
        "dataset": obj.dataset,
        "target_cols": obj.target_cols,
        "input_cols": obj.input_cols,
        "metrics": obj.metrics,
        "main_metric": obj.main_metric,
        "name": obj.name,
    }

    train_split = obj.split[0]
    test_split = obj.split[1]

    # One or more empty test partitions
    with pytest.raises(ValidationError):
        cls(split=(train_split,), **default_kwargs)
    with pytest.raises(ValidationError):
        cls(split=(train_split, []), **default_kwargs)
    with pytest.raises(ValidationError):
        cls(split=(train_split, {"test": []}), **default_kwargs)
    # Non-exclusive partitions
    with pytest.raises(ValidationError):
        cls(split=(train_split, test_split["test"] + train_split[:1]), **default_kwargs)
    with pytest.raises(ValidationError):
        cls(split=(train_split, {"test1": test_split, "test2": train_split[:1]}), **default_kwargs)
    # Invalid indices
    with pytest.raises(ValidationError):
        cls(split=(train_split + [len(obj.dataset)], test_split), **default_kwargs)
    with pytest.raises(ValidationError):
        cls(split=(train_split + [-1], test_split), **default_kwargs)
    with pytest.raises(ValidationError):
        cls(split=(train_split, test_split["test"] + [len(obj.dataset)]), **default_kwargs)
    with pytest.raises(ValidationError):
        cls(split=(train_split, test_split["test"] + [-1]), **default_kwargs)
    # Duplicate indices
    with pytest.raises(ValidationError):
        cls(split=(train_split + train_split[:1], test_split), **default_kwargs)
    with pytest.raises(ValidationError):
        cls(split=(train_split, test_split["test"] + test_split["test"][:1]), **default_kwargs)
    with pytest.raises(ValidationError):
        cls(
            split=(train_split, {"test1": test_split, "test2": test_split["test"] + test_split["test"][:1]}),
            **default_kwargs,
        )

    # It should _not_ fail with duplicate indices across test partitions
    cls(split=(train_split, {"test1": test_split["test"], "test2": test_split["test"]}), **default_kwargs)
    # It should _not_ fail with missing indices
    cls(split=(train_split[:-1], test_split), **default_kwargs)
    # It should _not_ fail with an empty train set
    benchmark = cls(split=([], test_split), **default_kwargs)
    train, _ = benchmark.get_train_test_split()
    assert len(train) == 0


@pytest.mark.parametrize("cls", [SingleTaskBenchmarkSpecification, MultiTaskBenchmarkSpecification])
def test_benchmark_column_verification(test_single_task_benchmark, test_multi_task_benchmark, cls):
    """Verifies that the column validation works as expected."""
    base = (
        test_single_task_benchmark if cls == SingleTaskBenchmarkSpecification else test_multi_task_benchmark
    )

    # By using the fixture as a default, we know it doesn't always fail
    default_kwargs = {
        "dataset": base.dataset,
        "metrics": base.metrics,
        "split": base.split,
        "name": base.name,
    }

    x_cols = base.input_cols
    y_cols = base.target_cols

    # Empty columns
    with pytest.raises(ValueError):
        cls(input_cols=[], target_cols=y_cols, **default_kwargs)
    with pytest.raises(ValueError):
        cls(input_cols=x_cols, target_cols=[], **default_kwargs)
    # Non-existent columns
    with pytest.raises(ValueError):
        cls(input_cols=["invalid", "smiles"], target_cols=y_cols, **default_kwargs)
    with pytest.raises(ValueError):
        cls(input_cols=x_cols, target_cols=["invalid", "expt"], **default_kwargs)


@pytest.mark.parametrize("cls", [SingleTaskBenchmarkSpecification, MultiTaskBenchmarkSpecification])
def test_benchmark_metrics_verification(test_single_task_benchmark, test_multi_task_benchmark, cls):
    """Verifies that the metric validation works as expected."""
    # By using the fixture as a default, we know it doesn't always fail
    base = (
        test_single_task_benchmark if cls == SingleTaskBenchmarkSpecification else test_multi_task_benchmark
    )

    default_kwargs = {
        "dataset": base.dataset,
        "target_cols": base.target_cols,
        "input_cols": base.input_cols,
        "split": base.split,
        "name": base.name,
    }

    # Invalid metric
    with pytest.raises(ValidationError):
        cls(metrics=["invalid"], **default_kwargs)
    with pytest.raises(ValidationError):
        cls(metrics="invalid", **default_kwargs)
    with pytest.raises(ValidationError):
        metrics_list = list(base.metrics)
        cls(
            metrics=metrics_list + [metrics_list[0]],
            **default_kwargs,
        )


def test_benchmark_from_json(test_single_task_benchmark, tmp_path):
    """Test whether we can successfully save and load a benchmark from JSON."""
    path = test_single_task_benchmark.to_json(str(tmp_path))
    new_benchmark = SingleTaskBenchmarkSpecification.from_json(path)
    assert new_benchmark == test_single_task_benchmark

    new_benchmark = load_benchmark(path)
    assert new_benchmark == test_single_task_benchmark


@pytest.mark.parametrize("fixture", ["test_single_task_benchmark", "test_multi_task_benchmark"])
def test_benchmark_checksum(fixture, request):
    """Test whether the checksum captures our notion of equality."""

    benchmark = request.getfixturevalue(fixture)
    cls = type(benchmark)
    metrics_list = list(benchmark.metrics)

    # Make sure the `md5sum` is part of the model dump even if not initiated yet.
    # This is important for uploads to the Hub.
    assert benchmark._md5sum is None and "md5sum" in benchmark.model_dump()

    original = benchmark.md5sum
    assert original is not None

    # --- Test that the checksum is the same with insignificant changes ---

    # Without any changes, same hash
    kwargs = benchmark.model_dump()
    assert cls(dataset=benchmark.dataset, **kwargs).md5sum == original

    # With a different ordering of the target columns
    kwargs["target_cols"] = kwargs["target_cols"][::-1]
    assert cls(dataset=benchmark.dataset, **kwargs).md5sum == original

    # With a different ordering of the metrics
    kwargs["metrics"] = metrics_list[::-1]
    assert cls(dataset=benchmark.dataset, **kwargs).md5sum == original

    # With a different ordering of the split
    kwargs["split"] = kwargs["split"][0][::-1], kwargs["split"][1]
    assert cls(dataset=benchmark.dataset, **kwargs).md5sum == original

    # --- Test that the checksum is NOT the same ---
    def _check_for_failure(_kwargs):
        assert cls(dataset=benchmark.dataset, **_kwargs).md5sum != _kwargs["md5sum"]

    # Split
    kwargs = benchmark.model_dump()
    kwargs["split"] = kwargs["split"][0][1:], kwargs["split"][1]
    _check_for_failure(kwargs)

    kwargs = benchmark.model_dump()
    kwargs["split"] = kwargs["split"][0], kwargs["split"][1]["test"][1:]
    _check_for_failure(kwargs)

    # Metrics
    kwargs = benchmark.model_dump()
    kwargs["metrics"] = kwargs["metrics"][1:] + ["accuracy"]
    kwargs["main_metric"] = kwargs["metrics"][0]
    _check_for_failure(kwargs)

    # Target columns
    kwargs = benchmark.model_dump()
    kwargs["target_cols"] = kwargs["target_cols"][1:] + ["iupac"]
    kwargs.pop("target_types", None)  # Reset target types that matches deleted target column
    _check_for_failure(kwargs)

    # Input columns
    kwargs = benchmark.model_dump()
    kwargs["input_cols"] = kwargs["input_cols"][1:] + ["iupac"]
    _check_for_failure(kwargs)

    # --- Don't fail if no checksum is provided ---
    kwargs["md5sum"] = None
    dataset = cls(dataset=benchmark.dataset, **kwargs)
    assert dataset.md5sum is not None


def test_setting_an_invalid_checksum(test_single_task_benchmark):
    """Test whether setting an invalid checksum raises an error."""
    with pytest.raises(ValueError):
        test_single_task_benchmark.md5sum = "invalid"


def test_checksum_verification(test_single_task_benchmark):
    """Test whether setting an invalid checksum raises an error."""
    test_single_task_benchmark.verify_checksum()
    test_single_task_benchmark.md5sum = "0" * 32
    with pytest.raises(ValueError):
        test_single_task_benchmark.verify_checksum()


def test_benchmark_duplicate_metrics(test_single_task_benchmark):
    """Tests that passing duplicate metrics will raise a validation error"""
    m = test_single_task_benchmark.model_dump()

    with pytest.raises(ValidationError, match="The benchmark specifies duplicate metric"):
        m["metrics"] = [
            Metric(label="roc_auc", config={"group_by": "CLASS_expt"}),
            Metric(label="roc_auc", config={"group_by": "CLASS_expt"}),
        ]
        m["main_metric"] = m["metrics"][0]
        SingleTaskBenchmarkSpecification(**m)

    with pytest.raises(ValidationError, match="The metrics of a benchmark need to have unique names."):
        m["metrics"][0].config.group_by = "MULTICLASS_calc"
        SingleTaskBenchmarkSpecification(**m)

    m["metrics"][0].custom_name = "custom_name"
    SingleTaskBenchmarkSpecification(dataset=test_single_task_benchmark.dataset, **m)


def test_benchmark_metric_deserialization(test_single_task_benchmark):
    """Tests that passing metrics as a list of strings or dictionaries works as expected"""
    m = test_single_task_benchmark.model_dump()

    # Should work with strings
    m["metrics"] = ["mean_absolute_error", "accuracy"]
    m["main_metric"] = "accuracy"
    SingleTaskBenchmarkSpecification(dataset=test_single_task_benchmark.dataset, **m)

    # Should work with dictionaries
    m["metrics"] = [
        {"label": "mean_absolute_error", "config": {"group_by": "CLASS_expt"}},
        {"label": "accuracy"},
    ]
    SingleTaskBenchmarkSpecification(dataset=test_single_task_benchmark.dataset, **m)



import json
from typing import Any, Callable, Literal, TypeAlias

import numpy as np
from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    PrivateAttr,
    computed_field,
    field_validator,
    model_validator,
)
from sklearn.metrics import (
    accuracy_score,
    balanced_accuracy_score,
    explained_variance_score,
    f1_score,
    matthews_corrcoef,
    mean_absolute_error,
    mean_squared_error,
    r2_score,
    roc_auc_score,
)
from typing_extensions import Self

from polaris.dataset._subset import Subset
from polaris.evaluate._predictions import BenchmarkPredictions
from polaris.evaluate.metrics import (
    absolute_average_fold_error,
    average_precision_score,
    cohen_kappa_score,
    pearsonr,
    spearman,
)
from polaris.evaluate.metrics.docking_metrics import rmsd_coverage
from polaris.utils.types import DirectionType, PredictionKwargs

GroundTruth: TypeAlias = Subset


def prepare_predictions(
    y_pred: BenchmarkPredictions,
    y_prob: BenchmarkPredictions,
    y_type: PredictionKwargs,
) -> BenchmarkPredictions:
    """
    Check that the correct type of predictions are passed to the metric.

    Args:
        y_pred: The predicted target values, if any.
        y_prob: The predicted target probabilities, if any.
        y_type: The type of predictions expected by the metric interface
    """

    if y_pred is None and y_prob is None:
        raise ValueError("Neither `y_pred` nor `y_prob` is specified.")

    if y_type == "y_pred":
        if y_pred is None:
            raise ValueError("Metric requires `y_pred` input.")
        pred = y_pred

    else:
        if y_prob is None:
            raise ValueError("Metric requires `y_prob` input.")
        pred = y_prob

    return pred


def mask_index(input_values):
    """
    Mask the NaN values in the input array

    Args:
        input_values: The input array to mask.
    """

    if np.issubdtype(input_values.dtype, np.number):
        mask = ~np.isnan(input_values)
    else:
        # Create a mask to identify NaNs
        mask = np.full(input_values.shape, True, dtype=bool)
        # Iterate over the array to identify NaNs
        for index, value in np.ndenumerate(input_values):
            # Any None value is NaN
            if value is None:
                mask[index] = False
    return mask


def prepare_metric_kwargs(
    y_true: np.ndarray,
    y_pred: np.ndarray,
    y_type: PredictionKwargs,
) -> BenchmarkPredictions:
    """
    Prepare the arguments for the metric function.

    Args:
        y_true: The true target values of shape (n_samples,)
        y_pred: The predicted target values of shape (n_samples,).
        y_type: The type of predictions expected by the metric interface.
    """
    mask = mask_index(y_true)
    return {"y_true": y_true[mask], y_type: y_pred[mask]}


MetricKind: TypeAlias = Literal["default", "grouped"]

MetricLabel: TypeAlias = Literal[
    "mean_absolute_error",
    "mean_squared_error",
    "r2",
    "pearsonr",
    "spearmanr",
    "explained_var",
    "absolute_average_fold_error",
    "accuracy",
    "balanced_accuracy",
    "mcc",
    "cohen_kappa",
    "pr_auc",
    "f1",
    "roc_auc",
    "f1_macro",
    "f1_micro",
    "roc_auc_ovr",
    "roc_auc_ovo",
    "rmsd_coverage",
]


class MetricInfo(BaseModel):
    """
    Metric metadata

    Attributes:
        fn: The callable that actually computes the metric.
        is_multitask: Whether the metric expects a single set of predictions or a dict of predictions.
        kwargs: Additional parameters required for the metric.
        direction: The direction for ranking of the metric,  "max" for maximization and "min" for minimization.
        y_type: The type of predictions expected by the metric interface.
    """

    fn: Callable
    is_multitask: bool = False
    kwargs: dict = Field(default_factory=dict)
    direction: DirectionType
    y_type: PredictionKwargs = "y_pred"

    model_config = ConfigDict(frozen=True)


DEFAULT_METRICS: dict[MetricLabel, MetricInfo] = {
    "mean_absolute_error": MetricInfo(fn=mean_absolute_error, direction="min"),
    "mean_squared_error": MetricInfo(fn=mean_squared_error, direction="min"),
    "r2": MetricInfo(fn=r2_score, direction="max"),
    "pearsonr": MetricInfo(fn=pearsonr, direction="max"),
    "spearmanr": MetricInfo(fn=spearman, direction="max"),
    "explained_var": MetricInfo(fn=explained_variance_score, direction="max"),
    "absolute_average_fold_error": MetricInfo(fn=absolute_average_fold_error, direction=1),
    "accuracy": MetricInfo(fn=accuracy_score, direction="max"),
    "balanced_accuracy": MetricInfo(fn=balanced_accuracy_score, direction="max"),
    "mcc": MetricInfo(fn=matthews_corrcoef, direction="max"),
    "cohen_kappa": MetricInfo(fn=cohen_kappa_score, direction="max"),
    "pr_auc": MetricInfo(fn=average_precision_score, direction="max", y_type="y_score"),
    "f1": MetricInfo(fn=f1_score, kwargs={"average": "binary"}, direction="max"),
    "roc_auc": MetricInfo(fn=roc_auc_score, direction="max", y_type="y_score"),
    "f1_macro": MetricInfo(fn=f1_score, kwargs={"average": "macro"}, direction="max"),
    "f1_micro": MetricInfo(fn=f1_score, kwargs={"average": "micro"}, direction="max"),
    "roc_auc_ovr": MetricInfo(
        fn=roc_auc_score, kwargs={"multi_class": "ovr"}, direction="max", y_type="y_score"
    ),
    "roc_auc_ovo": MetricInfo(
        fn=roc_auc_score, kwargs={"multi_class": "ovo"}, direction="max", y_type="y_score"
    ),
    "rmsd_coverage": MetricInfo(fn=rmsd_coverage, direction="max", y_type="y_pred"),
}


class GroupedMetricConfig(BaseModel):
    """
    The configuration for a GroupedMetric

    Attributes:
        group_by: The column to group by.
        on_error: How to handle errors when computing the metric.
        default: The default value to use when an error occurs.
        aggregation: The aggregation method to use when combining the metric scores.
    """

    _kind: MetricKind = PrivateAttr("grouped")

    group_by: str
    on_error: Literal["ignore", "raise", "default"] = "raise"
    default: float | None = None
    aggregation: Literal["mean", "median"] = "mean"

    @model_validator(mode="after")
    def check_default_required(self) -> Self:
        """Check that a default value is provided when on_error is 'default'."""
        handling = self.on_error
        default = self.default
        if handling == "default" and default is None:
            raise ValueError("Default value must be provided when on_error is 'default'")
        return self

    def score(
        self,
        y_true: GroundTruth,
        y_pred: BenchmarkPredictions | None,
        y_prob: BenchmarkPredictions | None,
        info: MetricInfo,
    ) -> float:
        """Compute the metric.

        Args:
            y_true: The true target values.
            y_pred: The predicted target values, if any.
            y_prob: The predicted target probabilities, if any.
            info: The metadata for the metric.
        """

        pred = prepare_predictions(y_pred, y_prob, info.y_type)

        # NOTE (cwognum): We rely on pandas to do the grouping.
        # This does imply a memory bottleneck.

        df = y_true.extend_inputs(self.group_by).as_dataframe()
        df[info.y_type] = pred.flatten()

        # Compute the metric for each group
        scores = []

        for _, group in df.groupby(self.group_by):
            y_true_group = group[y_true.target_cols[0]].values
            y_pred_group = group[info.y_type].values

            kwargs = prepare_metric_kwargs(
                y_true_group,
                y_pred_group,
                info.y_type,
            )

            # A group can be arbitrarily small (e.g. only 1 item), which also has implications
            # for its target distribution (e.g. only the positive class). These conditions can
            # lead to errors in the metric computation, which we handle here.
            try:
                score = info.fn(**kwargs, **info.kwargs)
            except Exception:
                if self.on_error == "ignore":
                    continue
                elif self.on_error == "raise":
                    raise
                score = self.default
            scores.append(score)

        # Aggregate the per-group score in a single score.
        if self.aggregation == "mean":
            score = np.mean(scores)
        else:
            score = np.median(scores)
        return score


class Metric(BaseModel):
    """
    A Metric in Polaris.

    A metric consists of a default metric, which is a callable labeled with additional metadata, as well as a config.
    The config can change how the metric is computed, for example by grouping the data before computing the metric.

    Attributes:
        label: The actual callable that is at the core of the metric implementation.
        custom_name: A optional, custom name of the metric. Names should be unique within the context of a benchmark.
        config: For more complex metrics, this object should hold all parameters for the metric.
        fn: The callable that actually computes the metric, automatically set based on the label.
        is_multitask: Whether the metric expects a single set of predictions or a dict of predictions, automatically set based on the label.
        kwargs: Additional parameters required for the metric, automatically set based on the label.
        direction: The direction for ranking of the metric,  "max" for maximization and "min" for minimization, automatically set based on the label.
        y_type: The type of predictions expected by the metric interface, automatically set based on the label.
    """

    label: MetricLabel
    config: GroupedMetricConfig | None = None
    custom_name: str | None = Field(None, exclude=True, alias="name")

    # Frozen metadata
    fn: Callable = Field(frozen=True, exclude=True)
    is_multitask: bool = Field(False, frozen=True, exclude=True)
    kwargs: dict = Field(default_factory=dict, frozen=True, exclude=True)
    direction: DirectionType = Field(frozen=True, exclude=True)
    y_type: PredictionKwargs = Field("y_pred", frozen=True, exclude=True)

    @model_validator(mode="before")
    @classmethod
    def populate_metric_info(cls, data: Any) -> Any:
        if isinstance(data, dict):
            label = data.get("label")
            metric_info = DEFAULT_METRICS.get(label)
            if metric_info is None:
                raise ValueError(f"Unknown metric label: {label}")
            data.update(metric_info)
        return data

    @field_validator("is_multitask")
    @classmethod
    def disable_multitask_metrics(cls, value):
        if value:
            raise NotImplementedError(
                "Multitask metrics are not yet supported and will require non-obvious changes to the Metric interface."
            )

    @computed_field
    @property
    def kind(self) -> MetricKind:
        return "default" if self.config is None else self.config._kind

    @computed_field
    @property
    def name(self) -> str:
        if self.custom_name is not None:
            return self.custom_name

        name = self.label
        if self.kind != "default":
            name += "_" + self.kind
        return name

    def _default_score(
        self,
        y_true: GroundTruth,
        y_pred: BenchmarkPredictions | None = None,
        y_prob: BenchmarkPredictions | None = None,
    ) -> float:
        """Compute the metric.

        Args:
            y_true: The true target values.
            y_pred: The predicted target values, if any.
            y_prob: The predicted target probabilities, if any.
        """
        pred = prepare_predictions(y_pred, y_prob, self.y_type)
        kwargs = prepare_metric_kwargs(
            y_true.as_array("y"),
            pred.flatten(),
            self.y_type,
        )
        return self.fn(**kwargs, **self.kwargs)

    def score(
        self,
        y_true: GroundTruth,
        y_pred: BenchmarkPredictions | None = None,
        y_prob: BenchmarkPredictions | None = None,
    ) -> float:
        """Compute the metric.

        Args:
            y_true: The true target values.
            y_pred: The predicted target values, if any.
            y_prob: The predicted target probabilities, if any.
        """
        if self.kind == "default":
            return self._default_score(y_true, y_pred, y_prob)
        return self.config.score(y_true, y_pred, y_prob, DEFAULT_METRICS[self.label])

    def __call__(
        self,
        y_true: GroundTruth,
        y_pred: BenchmarkPredictions | None = None,
        y_prob: BenchmarkPredictions | None = None,
    ) -> float:
        """
        For convenience, calling a `Metric` is the same as calling its `score` method.

        ```python
        metric = Metric(label="mean_absolute_error")
        assert metric.score(y_true=first, y_pred=second) == metric(y_true=first, y_pred=second)
        ```
        """
        return self.score(y_true, y_pred, y_prob)

    def __hash__(self) -> int:
        return hash(str(self))

    def __eq__(self, other) -> bool:
        match other:
            case Metric():
                return hash(other) == hash(self)
            case _:
                return False

    def __str__(self) -> str:
        return self.label + (
            json.dumps(self.config.model_dump(mode="json"), sort_keys=True) if self.config else ""
        )



from copy import copy
from typing import Callable, Iterable, Literal, Sequence

import numpy as np
import pandas as pd
from typing_extensions import Self

from polaris.dataset import DatasetV1
from polaris.dataset._adapters import Adapter
from polaris.utils.errors import TestAccessError
from polaris.utils.types import DatapointType


class Subset:
    """The `Subset` class provides easy access to a single partition of a split dataset.

    Info: No need to create this class manually
        You should not have to create this class manually. In most use-cases, you can create a `Subset` through the
        `get_train_test_split` method of a `BenchmarkSpecification` object.

    Tip: Featurize your inputs
        Not all datasets are already featurized. For example, a small-molecule task might simply provide the SMILES string.
        To easily featurize the inputs, you can pass or set a transformation function. For example:

        ```python
        import datamol as dm

        benchmark.get_train_test_split(..., featurization_fn=dm.to_fp)
        ```

    This should be the starting point for any framework-specific (e.g. PyTorch, Tensorflow) data-loader implementation.
    How the data is loaded in Polaris can be non-trivial, so this class is provided to abstract away the details.
    To easily build framework-specific data-loaders, a `Subset` supports various styles of accessing the data:

    1. **In memory**: Loads the entire dataset in memory and returns a single array with all datapoints,
        this style is accessible through the `subset.targets` and `subset.inputs` properties.
    2. **List**: Index the subset like a list, this style is accessible through the `subset[idx]` syntax.
    3. **Iterator**: Iterate over the subset, this style is accessible through the `iter(subset)` syntax.

    Examples:
        The different styles of accessing the data:

        ```python
        import polaris as po

        benchmark = po.load_benchmark(...)
        train, test = benchmark.get_train_test_split()

        # Load the entire dataset in memory, useful for e.g. scikit-learn.
        X = train.inputs
        y = train.targets

        # Access a single datapoint as with a list, useful for e.g. PyTorch.
        x, y = train[0]

        # Iterate over the dataset, useful for very large datasets.
        for x, y in train:
            ...
        ```

    Raises:
        TestAccessError: When trying to access the targets of the test set (specified by the `hide_targets` attribute).
    """

    def __init__(
        self,
        dataset: DatasetV1,
        indices: list[int | Sequence[int]],
        input_cols: Iterable[str] | str,
        target_cols: Iterable[str] | str,
        adapters: dict[str, Adapter] | None = None,
        featurization_fn: Callable | None = None,
        hide_targets: bool = False,
    ):
        self.dataset = dataset
        self.indices = indices
        self.target_cols = [target_cols] if isinstance(target_cols, str) else list(target_cols)
        self.input_cols = [input_cols] if isinstance(input_cols, str) else list(input_cols)

        self._adapters = adapters
        self._featurization_fn = featurization_fn

        # Storing all indices in memory can be memory consuming for XL datasets.
        # This is why we constrain the iloc to loc mapping to be the identity function for Dataset V2.
        match self.dataset:
            case DatasetV1():
                self._iloc_to_loc = lambda idx: self.dataset.rows[idx]
            case _:
                self._iloc_to_loc = lambda idx: idx

        # For the iterator implementation
        self._pointer = 0

        # This is a protected attribute to make explicit it should not be changed once set.
        self._hide_targets = hide_targets

    @property
    def is_multi_task(self):
        return len(self.target_cols) > 1

    @property
    def is_multi_input(self):
        return len(self.input_cols) > 1

    @property
    def inputs(self):
        """Alias for `self.as_array("x")`"""
        return self.as_array("x")

    @property
    def X(self):
        """Alias for `self.as_array("x")`"""
        return self.as_array("x")

    @property
    def targets(self):
        """Alias for `self.as_array("y")`"""
        return self.as_array("y")

    @property
    def y(self):
        """Alias for `self.as_array("y")`"""
        return self.as_array("y")

    def _get_single(
        self,
        row: str | int,
        cols: list[str],
        featurization_fn: Callable | None,
    ):
        """
        Loads a subset of the variables for a single data-point from the datasets.
        The dataset stores datapoint in a row-wise manner, so this method is used to access a single row.

        Args:
            row: The row index of the datapoint.
            cols: The columns (i.e. variables) to load for that data point.
            featurization_fn: The transformation function to apply to the data-point.
        """
        # Load the data-point
        # Also handles loading data stored in external files for pointer columns
        if len(cols) > 1:
            ret = {col: self.dataset.get_data(row, col, adapters=self._adapters) for col in cols}
        else:
            ret = self.dataset.get_data(row, cols[0], adapters=self._adapters)

        # Featurize
        if featurization_fn is not None:
            ret = featurization_fn(ret)

        return ret

    def _get_single_input(self, row: str | int):
        """Get a single input for a specific data-point and given the benchmark specification."""
        return self._get_single(row, self.input_cols, self._featurization_fn)

    def _get_single_output(self, row: str | int):
        """Get a single output for a specific data-point and given the benchmark specification."""
        return self._get_single(row, self.target_cols, None)

    def as_array(self, data_type: Literal["x", "y", "xy"]):
        """
        Scikit-learn style access to the targets and inputs.
        If the dataset is multi-target, this will return a dict of arrays.
        """

        if data_type == "xy":
            return self.as_array("x"), self.as_array("y")

        if data_type == "y" and self._hide_targets:
            raise TestAccessError("Within Polaris you should not need to access the targets of the test set")

        # We reset the index of the Pandas Table during Dataset class validation.
        # We can thus always assume that .iloc[idx] is the same as .loc[idx].
        if data_type == "x":
            ret = [self._get_single_input(self._iloc_to_loc(idx)) for idx in self.indices]
        else:
            ret = [self._get_single_output(self._iloc_to_loc(idx)) for idx in self.indices]

        if not ((self.is_multi_input and data_type == "x") or (self.is_multi_task and data_type == "y")):
            # If the target format is not a dict, we can just create the array directly.
            # With a single-task or single-input data point, this will be a 1D array.
            # With a multi-task or multi-input data point, this will be a 2D array.
            return np.array(ret)

        # If the return format is a dict, we want to convert
        # from an array of dicts to a dict of arrays.
        if data_type == "y":
            ret = {k: np.array([v[k] for v in ret]) for k in self.target_cols}
        elif data_type == "x":
            ret = {k: np.array([v[k] for v in ret]) for k in self.input_cols}

        return ret

    def as_dataframe(self) -> pd.DataFrame:
        """
        Returns the subset as a Pandas DataFrame.

        Warning: Memory usage
            This method loads the entire dataset in memory.
        """
        # Create an empty dataframe
        cols = self.input_cols
        if not self._hide_targets:
            cols += self.target_cols
        df = pd.DataFrame(columns=cols)

        # Fill the dataframe
        if not self._hide_targets:
            targets = self.targets
            if not self.is_multi_task:
                targets = {self.target_cols[0]: targets}

            for k in targets:
                df[k] = targets[k]

        inputs = self.inputs
        if not self.is_multi_input:
            inputs = {self.input_cols[0]: inputs}

        for k in inputs:
            df[k] = inputs[k]

        return df

    def copy(self) -> Self:
        """Returns a copy of the subset."""
        return copy(self)

    def extend_inputs(self, input_cols: Iterable[str] | str) -> Self:
        """
        Extend the subset to include additional input columns.

        Args:
            input_cols: The input columns to add.
        """
        input_cols = [input_cols] if isinstance(input_cols, str) else list(input_cols)
        copy = self.copy()
        copy.input_cols = list(set(self.input_cols + input_cols))
        return copy

    def filter_targets(self, target_cols: list[str] | str) -> Self:
        """
        Filter the subset to only include the specified target columns.

        Args:
            target_cols: The target columns to keep.
        """
        target_cols_subset = target_cols if isinstance(target_cols, list) else [target_cols]

        # Verify all target columns are in the original subset
        if not all(col in self.target_cols for col in target_cols_subset):
            raise ValueError("All new target columns need to be in the original subset.")
        target_cols_subset = [c for c in self.target_cols if c in target_cols_subset]

        copy = self.copy()
        copy.target_cols = target_cols_subset

        return copy

    def __len__(self):
        return len(self.indices)

    def __getitem__(self, item) -> DatapointType:
        """
        This method always returns an (X, y) tuple

        Some special cases:

        1. It supports multi-input datasets, in which case X is a dict with multiple values.
        2. It supports multi-task datasets, in which case y is a dict with multiple values.
        3. In case a dataset has a pointer column (i.e. a path to an external file with the actual data),
            this method also loads that data to memory.
        """

        idx = self.indices[item]
        idx = self._iloc_to_loc(idx)

        # Load the input modalities
        ins = self._get_single_input(idx)

        if self._hide_targets:
            # If we are not allowed to access the targets, we return the inputs only.
            # This is useful to make accidental access to the test set less likely.
            return ins

        # Retrieve the targets
        outs = self._get_single_output(idx)
        return ins, outs

    def __iter__(self):
        """Iterator implementation"""
        self._pointer = 0
        return self

    def __next__(self):
        """
        Allows for iterator-style access to the dataset, e.g. useful when datasets get large.
        Remember that pointer columns are not loaded into memory yet.
        """
        if self._pointer >= len(self):
            raise StopIteration

        item = self[self._pointer]
        self._pointer += 1
        return item



from polaris.evaluate.metrics.docking_metrics import rmsd_coverage
from polaris.evaluate.metrics.generic_metrics import (
    absolute_average_fold_error,
    average_precision_score,
    cohen_kappa_score,
    pearsonr,
    spearman,
)



from contextlib import contextmanager
from contextvars import ContextVar
from itertools import cycle

from rich.progress import (
    BarColumn,
    MofNCompleteColumn,
    Progress,
    SpinnerColumn,
    TextColumn,
    TimeElapsedColumn,
)

# Singleton Progress instance to be used for all calls to `track_progress`
progress_instance = ContextVar(
    "progress",
    default=Progress(
        SpinnerColumn(),
        TextColumn("[progress.description]{task.description}"),
        BarColumn(),
        MofNCompleteColumn(),
        TimeElapsedColumn(),
    ),
)

colors = cycle(
    {
        "green",
        "cyan",
        "magenta",
    }
)


@contextmanager
def track_progress(description: str, total: float | None = 1.0):
    """
    Use the Progress instance to track a task's progress
    """
    progress = progress_instance.get()

    # Make sure the Progress is started
    progress.start()

    task = progress.add_task(f"[{next(colors)}]{description}", total=total)

    try:
        # Yield the task and Progress instance, for more granular control
        yield progress, task

        # Mark the task as completed
        progress.update(task, completed=total, refresh=True)
        progress.log(f"[green] Success: {description}")
    except Exception:
        progress.log(f"[red] Error: {description}")
        raise
    finally:
        # Remove the task from the UI, and stop the progress bar if all tasks are completed
        progress.remove_task(task)
        if progress.finished:
            progress.stop()



from ._checksum import ZarrFileChecksum, compute_zarr_checksum
from ._manifest import generate_zarr_manifest
from ._memmap import MemoryMappedDirectoryStore

__all__ = [
    "MemoryMappedDirectoryStore",
    "compute_zarr_checksum",
    "ZarrFileChecksum",
    "generate_zarr_manifest",
]



import os

import boto3
import pytest
from moto import mock_aws

from polaris.hub.storage import S3Store


@pytest.fixture(scope="function")
def aws_credentials():
    """
    Mocked AWS Credentials for moto.
    """
    os.environ["AWS_ACCESS_KEY_ID"] = "testing"
    os.environ["AWS_SECRET_ACCESS_KEY"] = "testing"
    os.environ["AWS_SECURITY_TOKEN"] = "testing"
    os.environ["AWS_SESSION_TOKEN"] = "testing"


@pytest.fixture(scope="function")
def mocked_aws(aws_credentials):
    """
    Mock all AWS interactions
    Requires you to create your own boto3 clients
    """
    with mock_aws():
        yield


@pytest.fixture
def s3_store(mocked_aws):
    # Setup mock S3 environment
    s3 = boto3.client("s3", region_name="us-east-1")
    bucket_name = "test-bucket"
    s3.create_bucket(Bucket=bucket_name)

    # Create an instance of your S3Store
    store = S3Store(
        path=f"{bucket_name}/prefix",
        access_key="fake-access-key",
        secret_key="fake-secret-key",
        token="fake-token",
        endpoint_url="https://s3.amazonaws.com",
    )

    yield store


def test_set_and_get_item(s3_store):
    key = "test-key"
    value = b"test-value"
    s3_store[key] = value

    retrieved_value = s3_store[key]
    assert retrieved_value == value


def test_get_nonexistent_item(s3_store):
    with pytest.raises(KeyError):
        _ = s3_store["nonexistent-key"]


def test_contains_item(s3_store):
    key = "test-key"
    value = b"test-value"
    s3_store[key] = value

    assert key in s3_store
    assert "nonexistent-key" not in s3_store


def test_store_iterator_empty(s3_store):
    stored_keys = list(s3_store)
    assert stored_keys == []


def test_store_iterator(s3_store):
    keys = ["dir1/subdir1", "dir1/subdir2", "dir1/file1.ext", "dir2/file2.ext"]
    for key in keys:
        s3_store[key] = b"test"

    stored_keys = list(s3_store)
    assert sorted(stored_keys) == sorted(keys)


def test_store_length(s3_store):
    keys = ["dir1/subdir1", "dir1/subdir2", "dir1/file1.ext", "dir2/file2.ext"]
    for key in keys:
        s3_store[key] = b"test"

    assert len(s3_store) == len(keys)


def test_listdir(s3_store):
    keys = ["dir1/subdir1", "dir1/subdir2", "dir1/file1.ext", "dir2/file2.ext"]
    for key in keys:
        s3_store[key] = b"test"

    dir1_contents = list(s3_store.listdir("dir1"))
    assert set(dir1_contents) == {"file1.ext", "subdir1", "subdir2"}

    dir1_contents = list(s3_store.listdir())
    assert set(dir1_contents) == {"dir1", "dir2"}


def test_getsize(s3_store):
    key = "test-key"
    value = b"test-value"
    s3_store[key] = value

    size = s3_store.getsize(key)
    assert size == len(value)


def test_getitems(s3_store):
    keys = ["dir1/subdir1", "dir1/subdir2", "dir1/file1.ext", "dir2/file2.ext"]
    for key in keys:
        s3_store[key] = b"test"

    items = s3_store.getitems(keys, contexts={})
    assert len(items) == len(keys)
    assert all(key in items for key in keys)


def test_delete_item_not_supported(s3_store):
    with pytest.raises(NotImplementedError):
        del s3_store["some-key"]



import json
import logging
from hashlib import md5
from io import BytesIO
from typing import get_args
from urllib.parse import urljoin

import httpx
import pandas as pd
import zarr
from authlib.integrations.base_client.errors import InvalidTokenError, MissingTokenError
from authlib.integrations.httpx_client import OAuth2Client, OAuthError
from authlib.oauth2 import OAuth2Error, TokenAuth
from authlib.oauth2.rfc6749 import OAuth2Token
from httpx import ConnectError, HTTPStatusError, Response
from typing_extensions import Self

from polaris.benchmark import (
    BenchmarkV1Specification,
    MultiTaskBenchmarkSpecification,
    SingleTaskBenchmarkSpecification,
)
from polaris.benchmark._benchmark_v2 import BenchmarkV2Specification
from polaris.competition import CompetitionSpecification
from polaris.model import Model
from polaris.dataset import Dataset, DatasetV1, DatasetV2
from polaris.evaluate import BenchmarkResultsV1, BenchmarkResultsV2, CompetitionPredictions
from polaris.hub.external_client import ExternalAuthClient
from polaris.hub.oauth import CachedTokenAuth
from polaris.hub.settings import PolarisHubSettings
from polaris.hub.storage import StorageSession
from polaris.utils.context import track_progress
from polaris.utils.errors import (
    InvalidDatasetError,
    PolarisCreateArtifactError,
    PolarisHubError,
    PolarisRetrieveArtifactError,
    PolarisSSLError,
    PolarisUnauthorizedError,
)
from polaris.utils.types import (
    AccessType,
    ChecksumStrategy,
    HubOwner,
    SupportedLicenseType,
    TimeoutTypes,
    ZarrConflictResolution,
)

logger = logging.getLogger(__name__)

_HTTPX_SSL_ERROR_CODE = "[SSL: CERTIFICATE_VERIFY_FAILED]"


class PolarisHubClient(OAuth2Client):
    """
    A client for the Polaris Hub API. The Polaris Hub is a central repository of datasets, benchmarks and results.
    Visit it here: [https://polarishub.io/](https://polarishub.io/).

    Bases the [`authlib` client](https://docs.authlib.org/en/latest/client/api.html#authlib.integrations.httpx_client.OAuth2Client),
    which in turns bases the [`httpx` client](https://www.python-httpx.org/advanced/#client-instances).
    See the relevant docs to learn more about how to use these clients outside of the integration with the Polaris Hub.

    Note: Closing the client
        The client should be closed after all requests have been made. For convenience, you can also use the
        client as a context manager to automatically close the client when the context is exited. Note that once
        the client has been closed, it cannot be used anymore.

        ```python
        # Make sure to close the client once finished
        client = PolarisHubClient()
        client.get(...)
        client.close()

        # Or use the client as a context manager
        with PolarisHubClient() as client:
            client.get(...)
        ```

    Note: Interacting with artifacts owned by an organization
        Soon after being added to a new organization on Polaris, there may be a delay spanning some
        minutes where you cannot upload/download artifacts where the aforementioned organization is the owner.
        If this occurs, please re-login via `polaris login --overwrite` and try again.

    Info: Async Client
        `authlib` also supports an [async client](https://docs.authlib.org/en/latest/client/httpx.html#async-oauth-2-0).
        Since we don't expect to make multiple requests to the Hub in parallel
        and due to the added complexity stemming from using the Python asyncio API,
        we are sticking to the sync client - at least for now.
    """

    def __init__(
        self,
        settings: PolarisHubSettings | None = None,
        cache_auth_token: bool = True,
        **kwargs: dict,
    ):
        """
        Args:
            settings: A `PolarisHubSettings` instance.
            cache_auth_token: Whether to cache the auth token to a file.
            **kwargs: Additional keyword arguments passed to the authlib `OAuth2Client` constructor.
        """
        self.settings = PolarisHubSettings() if settings is None else settings

        # We cache the auth token by default, but allow the user to disable this.
        self.token_auth_class = CachedTokenAuth if cache_auth_token else TokenAuth

        super().__init__(
            # OAuth2Client
            token_endpoint=self.settings.hub_token_url,
            token_endpoint_auth_method="none",
            # httpx.Client
            base_url=self.settings.api_url,
            cert=self.settings.ca_bundle,
            timeout=self.settings.default_timeout,
            # Extra
            **kwargs,
        )

        # We use an external client to get an auth token that can be exchanged for a Polaris Hub token
        self.external_client = ExternalAuthClient(
            settings=self.settings, cache_auth_token=cache_auth_token, **kwargs
        )

    def __enter__(self: Self) -> Self:
        """
        When used as a context manager, automatically check that authentication is valid.
        """
        super().__enter__()
        if not self.ensure_active_token():
            raise PolarisUnauthorizedError()
        return self

    @property
    def has_user_password(self) -> bool:
        return bool(self.settings.username and self.settings.password)

    def _prepare_token_endpoint_body(self, body, grant_type, **kwargs):
        """
        Override to support required fields for the token exchange grant type.
        See https://datatracker.ietf.org/doc/html/rfc8693#name-request
        """
        if grant_type == "urn:ietf:params:oauth:grant-type:token-exchange":
            kwargs.update(
                {
                    "subject_token": self.external_client.token["access_token"],
                    "subject_token_type": "urn:ietf:params:oauth:token-type:access_token",
                    "requested_token_type": "urn:ietf:params:oauth:token-type:jwt",
                }
            )
        return super()._prepare_token_endpoint_body(body, grant_type, **kwargs)

    def ensure_active_token(self, token: OAuth2Token | None = None) -> bool:
        """
        Override the active check to trigger a refetch of the token if it is not active.
        """
        # This won't be needed with if we set a lower bound for authlib: >=1.3.2
        # See https://github.com/lepture/authlib/pull/625
        # As of now, this latest version is not available on Conda though.
        token = token or self.token
        is_active = super().ensure_active_token(token) if token else False
        if is_active:
            return True

        # Check if external token is still valid, or we're using password auth
        if not (self.has_user_password or self.external_client.ensure_active_token()):
            return False

        # If so, use it to get a new Hub token
        self.token = self.fetch_token()
        return True

    def fetch_token(self, **kwargs):
        """
        Handles the optional support for password grant type, and provide better error messages.
        """
        try:
            return super().fetch_token(
                username=self.settings.username,
                password=self.settings.password,
                grant_type="password"
                if self.has_user_password
                else "urn:ietf:params:oauth:grant-type:token-exchange",
                **kwargs,
            )
        except (OAuthError, OAuth2Error) as error:
            raise PolarisHubError(
                message=f"Could not obtain a token to access the Hub. Error was: {error.error} - {error.description}"
            ) from error

    def _base_request_to_hub(self, url: str, method: str, **kwargs) -> Response:
        """Utility function since most API methods follow the same pattern"""
        try:
            response = self.request(url=url, method=method, **kwargs)
            response.raise_for_status()
            return response
        except HTTPStatusError as error:
            # If JSON is included in the response body, we retrieve it and format it for output. If not, we fall back to
            # retrieving plain text from the body. 500 errors will not have a JSON response.
            try:
                response_text = error.response.json()
                response_text = json.dumps(response_text, indent=2, sort_keys=True)
            except (json.JSONDecodeError, TypeError):
                response_text = error.response.text

            match error.response.status_code:
                case 401:
                    # Providing the user a more helpful error message suggesting a re-login when their access credentials expire.
                    raise PolarisUnauthorizedError(response_text=response_text) from error
                case 403:
                    # The error can happen due to the JWT token containing outdated information.
                    raise PolarisCreateArtifactError(response_text=response_text) from error
                case 404:
                    # This happens when an artifact doesn't exist _or_ when the user has no access to that artifact.
                    raise PolarisRetrieveArtifactError(response_text=response_text) from error
                case _:
                    raise PolarisHubError(response_text=response_text) from error

    def get_metadata_from_response(self, response: Response, key: str) -> str | None:
        """Get custom metadata saved to the R2 object from the headers."""
        key = f"{self.settings.custom_metadata_prefix}{key}"
        return response.headers.get(key)

    def request(self, method, url, withhold_token=False, auth=httpx.USE_CLIENT_DEFAULT, **kwargs):
        """Wraps the base request method to handle errors"""
        try:
            return super().request(method, url, withhold_token, auth, **kwargs)
        except ConnectError as error:
            # NOTE (cwognum): In the stack trace, the more specific SSLCertVerificationError is raised.
            #   We could use the traceback module to catch this specific error, but that would be overkill here.
            error_string = str(error)
            ExceptionClass = PolarisSSLError if _HTTPX_SSL_ERROR_CODE in error_string else PolarisHubError
            raise ExceptionClass(error_string) from error
        except (MissingTokenError, InvalidTokenError, OAuthError) as error:
            raise PolarisUnauthorizedError() from error

    def login(self, overwrite: bool = False, auto_open_browser: bool = True):
        """Login to the Polaris Hub using the OAuth2 protocol.

        Warning: Headless authentication
            It is currently not possible to login to the Polaris Hub without a browser.
            See [this Github issue](https://github.com/polaris-hub/polaris/issues/30) for more info.

        Args:
            overwrite: Whether to overwrite the current token if the user is already logged in.
            auto_open_browser: Whether to automatically open the browser to visit the authorization URL.
        """
        if overwrite or self.token is None or not self.ensure_active_token():
            self.external_client.interactive_login(overwrite=overwrite, auto_open_browser=auto_open_browser)
            self.token = self.fetch_token()

        logger.info("You are successfully logged in to the Polaris Hub.")

    # =========================
    #     API Endpoints
    # =========================

    def list_datasets(self, limit: int = 100, offset: int = 0) -> list[str]:
        """List all available datasets (v1 and v2) on the Polaris Hub.
        We prioritize v2 datasets over v1 datasets.

        Args:
            limit: The maximum number of datasets to return.
            offset: The offset from which to start returning datasets.

        Returns:
            A list of dataset names in the format `owner/dataset_slug`.
        """
        with track_progress(description="Fetching datasets", total=1):
            # Step 1: Fetch enough v2 datasets to cover the offset and limit
            v2_json_response = self._base_request_to_hub(
                url="/v2/dataset", method="GET", params={"limit": limit, "offset": offset}
            ).json()
            v2_data = v2_json_response["data"]
            v2_datasets = [dataset["artifactId"] for dataset in v2_data]

            # If v2 datasets satisfy the limit, return them
            if len(v2_datasets) == limit:
                return v2_datasets

            # Step 2: Calculate the remaining limit and fetch v1 datasets
            remaining_limit = max(0, limit - len(v2_datasets))

            v1_json_response = self._base_request_to_hub(
                url="/v1/dataset",
                method="GET",
                params={
                    "limit": remaining_limit,
                    "offset": max(0, offset - v2_json_response["metadata"]["total"]),
                },
            ).json()
            v1_data = v1_json_response["data"]
            v1_datasets = [dataset["artifactId"] for dataset in v1_data]

            # Combine the v2 and v1 datasets
            combined_datasets = v2_datasets + v1_datasets

            # Ensure the final combined list respects the limit
            return combined_datasets

    def get_dataset(
        self,
        owner: str | HubOwner,
        slug: str,
        verify_checksum: ChecksumStrategy = "verify_unless_zarr",
    ) -> DatasetV1 | DatasetV2:
        """Load a standard dataset from the Polaris Hub.

        Args:
            owner: The owner of the dataset. Can be either a user or organization from the Polaris Hub.
            slug: The slug of the dataset.
            verify_checksum: Whether to use the checksum to verify the integrity of the dataset. If None,
                will infer a practical default based on the dataset's storage location.

        Returns:
            A `Dataset` instance, if it exists.
        """
        with track_progress(description="Fetching dataset", total=1):
            try:
                return self._get_v1_dataset(owner, slug, verify_checksum)
            except PolarisRetrieveArtifactError:
                # If the v1 dataset is not found, try to load a v2 dataset
                return self._get_v2_dataset(owner, slug)

    def _get_v1_dataset(
        self,
        owner: str | HubOwner,
        slug: str,
        verify_checksum: ChecksumStrategy = "verify_unless_zarr",
    ) -> DatasetV1:
        """Loads a V1 dataset from Polaris Hub

        Args:
            owner: The owner of the dataset. Can be either a user or organization from the Polaris Hub.
            slug: The slug of the dataset.
            verify_checksum: Whether to use the checksum to verify the integrity of the dataset.

        Returns:
            A `Dataset` instance, if it exists.
        """
        url = f"/v1/dataset/{owner}/{slug}"
        response = self._base_request_to_hub(url=url, method="GET")
        response_data = response.json()

        # Disregard the Zarr root in the response. We'll get it from the storage token instead.
        response_data.pop("zarrRootPath", None)

        # Load the dataset table and optional Zarr archive
        with StorageSession(self, "read", Dataset.urn_for(owner, slug)) as storage:
            table = pd.read_parquet(BytesIO(storage.get_file("root")))
            zarr_root_path = storage.paths.extension

            if zarr_root_path is not None:
                # For V1 datasets, the Zarr Root is optional.
                # It should be None if the dataset does not use pointer columns
                zarr_root_path = str(zarr_root_path)

        dataset = DatasetV1(table=table, zarr_root_path=zarr_root_path, **response_data)
        md5sum = response_data["md5Sum"]

        if dataset.should_verify_checksum(verify_checksum):
            dataset.verify_checksum(md5sum)
        else:
            dataset.md5sum = md5sum

        return dataset

    def _get_v2_dataset(self, owner: str | HubOwner, slug: str) -> DatasetV2:
        """"""
        url = f"/v2/dataset/{owner}/{slug}"
        response = self._base_request_to_hub(url=url, method="GET")
        response_data = response.json()

        # Disregard the Zarr root in the response. We'll get it from the storage token instead.
        response_data.pop("zarrRootPath", None)

        # Load the Zarr archive
        with StorageSession(self, "read", DatasetV2.urn_for(owner, slug)) as storage:
            zarr_root_path = str(storage.paths.root)

        dataset = DatasetV2(zarr_root_path=zarr_root_path, **response_data)
        return dataset

    def list_benchmarks(self, limit: int = 100, offset: int = 0) -> list[str]:
        """List all available benchmarks (v1 and v2) on the Polaris Hub.
        We prioritize v2 benchmarks over v1 benchmarks.

        Args:
            limit: The maximum number of benchmarks to return.
            offset: The offset from which to start returning benchmarks.

        Returns:
            A list of benchmark names in the format `owner/benchmark_slug`.
        """
        with track_progress(description="Fetching benchmarks", total=1):
            # Step 1: Fetch enough v2 benchmarks to cover the offset and limit
            v2_json_response = self._base_request_to_hub(
                url="/v2/benchmark", method="GET", params={"limit": limit, "offset": offset}
            ).json()
            v2_data = v2_json_response["data"]
            v2_benchmarks = [benchmark["artifactId"] for benchmark in v2_data]

            # If v2 benchmarks satisfy the limit, return them
            if len(v2_benchmarks) == limit:
                return v2_benchmarks

            # Step 2: Calculate the remaining limit and fetch v1 benchmarks
            remaining_limit = max(0, limit - len(v2_benchmarks))

            v1_json_response = self._base_request_to_hub(
                url="/v1/benchmark",
                method="GET",
                params={
                    "limit": remaining_limit,
                    "offset": max(0, offset - v2_json_response["metadata"]["total"]),
                },
            ).json()
            v1_data = v1_json_response["data"]
            v1_benchmarks = [benchmark["artifactId"] for benchmark in v1_data]

            # Combine the v2 and v1 benchmarks
            combined_benchmarks = v2_benchmarks + v1_benchmarks

            # Ensure the final combined list respects the limit
            return combined_benchmarks

    def get_benchmark(
        self,
        owner: str | HubOwner,
        slug: str,
        verify_checksum: ChecksumStrategy = "verify_unless_zarr",
    ) -> BenchmarkV1Specification | BenchmarkV2Specification:
        """Load a benchmark from the Polaris Hub.

        Args:
            owner: The owner of the benchmark. Can be either a user or organization from the Polaris Hub.
            slug: The slug of the benchmark.
            verify_checksum: Whether to use the checksum to verify the integrity of the benchmark.

        Returns:
            A `BenchmarkSpecification` instance, if it exists.
        """
        with track_progress(description="Fetching benchmark", total=1):
            try:
                return self._get_v1_benchmark(owner, slug, verify_checksum)
            except PolarisRetrieveArtifactError:
                # If the v1 benchmark is not found, try to load a v2 benchmark
                return self._get_v2_benchmark(owner, slug)

    def _get_v1_benchmark(
        self,
        owner: str | HubOwner,
        slug: str,
        verify_checksum: ChecksumStrategy = "verify_unless_zarr",
    ) -> BenchmarkV1Specification:
        response = self._base_request_to_hub(url=f"/v1/benchmark/{owner}/{slug}", method="GET")
        response_data = response.json()

        # TODO (jstlaurent): response["dataset"]["artifactId"] is the owner/name unique identifier,
        #  but we'd need to change the signature of get_dataset to use it
        response_data["dataset"] = self.get_dataset(
            response_data["dataset"]["owner"]["slug"],
            response_data["dataset"]["name"],
            verify_checksum=verify_checksum,
        )

        # TODO (cwognum): As we get more complicated benchmarks, how do we still find the right subclass?
        #  Maybe through structural pattern matching, introduced in Py3.10, or Pydantic's discriminated unions?
        benchmark_cls = (
            SingleTaskBenchmarkSpecification
            if len(response_data["targetCols"]) == 1
            else MultiTaskBenchmarkSpecification
        )

        benchmark = benchmark_cls(**response_data)

        if benchmark.dataset.should_verify_checksum(verify_checksum):
            benchmark.verify_checksum()
        else:
            benchmark.md5sum = response_data["md5Sum"]

        return benchmark

    def _get_v2_benchmark(self, owner: str | HubOwner, slug: str) -> BenchmarkV2Specification:
        response = self._base_request_to_hub(url=f"/v2/benchmark/{owner}/{slug}", method="GET")
        response_data = response.json()

        response_data["dataset"] = self.get_dataset(*response_data["dataset"]["artifactId"].split("/"))

        # Load the split index sets
        with StorageSession(self, "read", BenchmarkV2Specification.urn_for(owner, slug)) as storage:
            split = {label: storage.get_file(label) for label in response_data.get("split", {}).keys()}

        return BenchmarkV2Specification(**{**response_data, "split": split})

    def upload_results(
        self,
        results: BenchmarkResultsV1 | BenchmarkResultsV2,
        access: AccessType = "private",
        owner: HubOwner | str | None = None,
    ):
        """Upload the results to the Polaris Hub.

        Info: Owner
            The owner of the results will automatically be inferred by the Hub from the user making the request.
            Contrary to other artifact types, an organization cannot own a set of results.
            However, you can specify the `BenchmarkResults.contributors` field to share credit with other hub users.

        Note: Required metadata
            The Polaris client and Hub maintain different requirements as to which metadata is required.
            The requirements by the Hub are stricter, so when uploading to the Hub you might
            get some errors on missing metadata. Make sure to fill-in as much of the metadata as possible
            before uploading.

        Args:
            results: The results to upload.
            access: Grant public or private access to result
            owner: Which Hub user or organization owns the artifact. Takes precedence over `results.owner`.
        """
        with track_progress(description="Uploading results", total=1) as (progress, task):
            # Get the serialized model data-structure
            results.owner = HubOwner.normalize(owner or results.owner)
            result_json = results.model_dump(by_alias=True, exclude_none=True)

            # Make a request to the Hub
            response = self._base_request_to_hub(
                url="/v2/result", method="POST", json={"access": access, **result_json}
            )

            # Inform the user about where to find their newly created artifact.
            result_url = urljoin(self.settings.hub_url, response.headers.get("Content-Location"))

            progress.log(
                f"[green]Your result has been successfully uploaded to the Hub. View it here: {result_url}"
            )

    def upload_dataset(
        self,
        dataset: DatasetV1 | DatasetV2,
        access: AccessType = "private",
        timeout: TimeoutTypes = (10, 200),
        owner: HubOwner | str | None = None,
        if_exists: ZarrConflictResolution = "replace",
        parent_artifact_id: str | None = None,
    ):
        """Upload a dataset to the Polaris Hub.

        Info: Owner
            You have to manually specify the owner in the dataset data model. Because the owner could
            be a user or an organization, we cannot automatically infer this from just the logged-in user.

        Note: Required metadata
            The Polaris client and Hub maintain different requirements as to which metadata is required.
            The requirements by the Hub are stricter, so when uploading to the Hub you might
            get some errors on missing metadata. Make sure to fill-in as much of the metadata as possible
            before uploading.

        Args:
            dataset: The dataset to upload.
            access: Grant public or private access to result
            timeout: Request timeout values. User can modify the value when uploading large dataset as needed.
                This can be a single value with the timeout in seconds for all IO operations, or a more granular
                tuple with (connect_timeout, write_timeout). The type of the the timout parameter comes from `httpx`.
                Since datasets can get large, it might be needed to increase the write timeout for larger datasets.
                See also: https://www.python-httpx.org/advanced/#timeout-configuration
            owner: Which Hub user or organization owns the artifact. Takes precedence over `dataset.owner`.
            if_exists: Action for handling existing files in the Zarr archive. Options are 'raise' to throw
                an error, 'replace' to overwrite, or 'skip' to proceed without altering the existing files.
            parent_artifact_id: The `owner/slug` of the parent dataset, if uploading a new version of a dataset.
        """
        # Normalize timeout
        if timeout is None:
            timeout = self.settings.default_timeout

        # Check if a dataset license was specified prior to upload
        if not dataset.license:
            raise InvalidDatasetError(
                f"\nPlease specify a supported license for this dataset prior to uploading to the Polaris Hub.\nOnly some licenses are supported - {get_args(SupportedLicenseType)}."
            )

        if isinstance(dataset, DatasetV1):
            self._upload_v1_dataset(dataset, timeout, access, owner, if_exists, parent_artifact_id)
        elif isinstance(dataset, DatasetV2):
            self._upload_v2_dataset(dataset, timeout, access, owner, if_exists, parent_artifact_id)

    def _upload_v1_dataset(
        self,
        dataset: DatasetV1,
        timeout: TimeoutTypes,
        access: AccessType,
        owner: HubOwner | str | None,
        if_exists: ZarrConflictResolution,
        parent_artifact_id: str | None,
    ):
        """
        Upload a V1 dataset to the Polaris Hub.
        """

        with track_progress(description="Uploading dataset", total=1) as (progress, task):
            # Get the serialized data-model
            # We exclude the table as it handled separately
            dataset.owner = HubOwner.normalize(owner or dataset.owner)
            dataset_json = dataset.model_dump(exclude={"table"}, exclude_none=True, by_alias=True)

            # If the dataset uses Zarr, we will save the Zarr archive to the Hub as well
            if dataset.uses_zarr:
                dataset_json["zarrRootPath"] = f"{StorageSession.polaris_protocol}://data.zarr"

            # Uploading a dataset is a three-step process.
            # 1. Upload the dataset metadata to the Hub and prepare the Hub to receive the data
            # 2. Upload the parquet file to Hub storage
            # 3. Upload the associated Zarr archive to Hub storage

            # Prepare the parquet file
            in_memory_parquet = BytesIO()
            dataset.table.to_parquet(in_memory_parquet)
            parquet_size = len(in_memory_parquet.getbuffer())
            parquet_md5 = md5(in_memory_parquet.getbuffer()).hexdigest()

            # Step 1: Upload metadata
            # Instead of directly uploading the data, we announce to the Hub that we intend to upload it.
            # We do so separately for the Zarr archive and Parquet file.
            url = f"/v1/dataset/{dataset.artifact_id}"
            response = self._base_request_to_hub(
                url=url,
                method="PUT",
                json={
                    "tableContent": {
                        "size": parquet_size,
                        "fileType": "parquet",
                        "md5Sum": parquet_md5,
                    },
                    "zarrContent": [md5sum.model_dump() for md5sum in dataset._zarr_md5sum_manifest],
                    "access": access,
                    "parentArtifactId": parent_artifact_id,
                    **dataset_json,
                },
                timeout=timeout,
            )

            inserted_dataset = response.json()

            # We modify the slug in the server
            # Update dataset.slug here so dataset.urn is constructed correctly
            dataset.slug = inserted_dataset["slug"]

            with StorageSession(self, "write", dataset.urn) as storage:
                with track_progress(description="Copying Parquet file", total=1) as (progress, task):
                    # Step 2: Upload the parquet file
                    progress.log("[yellow]This may take a while.")
                    storage.set_file("root", in_memory_parquet.getvalue())

                    # Step 3: Upload any associated Zarr archive
                if dataset.uses_zarr:
                    with track_progress(description="Copying Zarr archive", total=1):
                        destination = storage.store("extension")

                        # Locally consolidate Zarr archive metadata. Future updates on handling consolidated
                        # metadata based on Zarr developers' recommendations can be tracked at:
                        # https://github.com/zarr-developers/zarr-python/issues/1731
                        zarr.consolidate_metadata(dataset.zarr_root.store.store)
                        zmetadata_content = dataset.zarr_root.store.store[".zmetadata"]
                        destination[".zmetadata"] = zmetadata_content

                        # Copy the Zarr archive to the Hub
                        destination.copy_from_source(
                            dataset.zarr_root.store.store, if_exists=if_exists, log=logger.info
                        )

            dataset_url = urljoin(self.settings.hub_url, response.headers.get("Content-Location"))
            progress.log(
                f"[green]Your dataset has been successfully uploaded to the Hub.\nView it here: {dataset_url}"
            )

    def _upload_v2_dataset(
        self,
        dataset: DatasetV2,
        timeout: TimeoutTypes,
        access: AccessType,
        owner: HubOwner | str | None,
        if_exists: ZarrConflictResolution,
        parent_artifact_id: str | None,
    ):
        """
        Upload a V2 dataset to the Polaris Hub.
        """

        with track_progress(description="Uploading dataset", total=1) as (progress, task):
            # Get the serialized data-model
            dataset.owner = HubOwner.normalize(owner or dataset.owner)
            dataset_json = dataset.model_dump(exclude_none=True, by_alias=True)

            # Step 1: Upload dataset metadata
            url = f"/v2/dataset/{dataset.artifact_id}"
            response = self._base_request_to_hub(
                url=url,
                method="PUT",
                json={
                    "zarrManifestFileContent": {
                        "md5Sum": dataset.zarr_manifest_md5sum,
                    },
                    "access": access,
                    "parentArtifactId": parent_artifact_id,
                    **dataset_json,
                },
                timeout=timeout,
            )

            inserted_dataset = response.json()

            # We modify the slug in the server
            # Update dataset.slug here so dataset.urn is constructed correctly
            dataset.slug = inserted_dataset["slug"]

            with StorageSession(self, "write", dataset.urn) as storage:
                # Step 2: Upload the manifest file
                with track_progress(description="Copying manifest file", total=1):
                    with open(dataset.zarr_manifest_path, "rb") as manifest_file:
                        storage.set_file("manifest", manifest_file.read())

                # Step 3: Upload the Zarr archive
                with track_progress(description="Copying Zarr archive", total=1) as (
                    progress_zarr,
                    task_zarr,
                ):
                    progress_zarr.log("[yellow]This may take a while.")

                    destination = storage.store("root")

                    # Locally consolidate Zarr archive metadata. Future updates on handling consolidated
                    # metadata based on Zarr developers' recommendations can be tracked at:
                    # https://github.com/zarr-developers/zarr-python/issues/1731
                    zarr.consolidate_metadata(dataset.zarr_root.store.store)
                    zmetadata_content = dataset.zarr_root.store.store[".zmetadata"]
                    destination[".zmetadata"] = zmetadata_content

                    # Copy the Zarr archive to the Hub
                    destination.copy_from_source(
                        dataset.zarr_root.store.store, if_exists=if_exists, log=logger.info
                    )

            dataset_url = urljoin(self.settings.hub_url, response.headers.get("Content-Location"))
            progress.log(
                f"[green]Your V2 dataset has been successfully uploaded to the Hub.\nView it here: {dataset_url}"
            )

    def upload_benchmark(
        self,
        benchmark: BenchmarkV1Specification | BenchmarkV2Specification,
        access: AccessType = "private",
        owner: HubOwner | str | None = None,
        parent_artifact_id: str | None = None,
    ):
        """Upload a benchmark to the Polaris Hub.

        Info: Owner
            You have to manually specify the owner in the benchmark data model. Because the owner could
            be a user or an organization, we cannot automatically infer this from the logged-in user.

        Note: Required metadata
            The Polaris client and Hub maintain different requirements as to which metadata is required.
            The requirements by the Hub are stricter, so when uploading to the Hub you might
            get some errors on missing metadata. Make sure to fill-in as much of the metadata as possible
            before uploading.

        Note: Non-existent datasets
            The client will _not_ upload the associated dataset to the Hub if it does not yet exist.
            Make sure to specify an existing dataset or upload the dataset first.

        Args:
            benchmark: The benchmark to upload.
            access: Grant public or private access to result
            owner: Which Hub user or organization owns the artifact. Takes precedence over `benchmark.owner`.
            parent_artifact_id: The `owner/slug` of the parent benchmark, if uploading a new version of a benchmark.
        """
        match benchmark:
            case BenchmarkV1Specification():
                self._upload_v1_benchmark(benchmark, access, owner, parent_artifact_id)
            case BenchmarkV2Specification():
                self._upload_v2_benchmark(benchmark, access, owner, parent_artifact_id)

    def _upload_v1_benchmark(
        self,
        benchmark: BenchmarkV1Specification,
        access: AccessType = "private",
        owner: HubOwner | str | None = None,
        parent_artifact_id: str | None = None,
    ):
        """
        Upload a V1 benchmark to the Polaris Hub.
        """
        with track_progress(description="Uploading benchmark", total=1) as (progress, task):
            # Get the serialized data-model
            # We exclude the dataset as we expect it to exist on the Hub already.
            benchmark.owner = HubOwner.normalize(owner or benchmark.owner)
            benchmark_json = benchmark.model_dump(exclude={"dataset"}, exclude_none=True, by_alias=True)
            benchmark_json["datasetArtifactId"] = benchmark.dataset.artifact_id
            benchmark_json["access"] = access

            url = f"/v1/benchmark/{benchmark.artifact_id}"
            response = self._base_request_to_hub(
                url=url, method="PUT", json={"parentArtifactId": parent_artifact_id, **benchmark_json}
            )

            benchmark_url = urljoin(self.settings.hub_url, response.headers.get("Content-Location"))
            progress.log(
                f"[green]Your benchmark has been successfully uploaded to the Hub.\nView it here: {benchmark_url}"
            )

    def _upload_v2_benchmark(
        self,
        benchmark: BenchmarkV2Specification,
        access: AccessType = "private",
        owner: HubOwner | str | None = None,
        parent_artifact_id: str | None = None,
    ):
        """
        Upload a V2 benchmark to the Polaris Hub.
        """
        with track_progress(description="Uploading benchmark", total=1) as (progress, task):
            # Get the serialized data-model
            # We exclude the dataset as we expect it to exist on the Hub already.
            benchmark.owner = HubOwner.normalize(owner or benchmark.owner)
            benchmark_json = benchmark.model_dump(exclude_none=True, by_alias=True)

            # Uploading a V2 benchmark is a multistep process.
            # 1. Upload the benchmark metadata to the Hub and prepare the Hub to receive the data
            # 2. Upload each index set bitmap to the Hub storage

            # Step 1: Upload metadata
            url = f"/v2/benchmark/{benchmark.artifact_id}"
            response = self._base_request_to_hub(
                url=url,
                method="PUT",
                json={
                    "access": access,
                    "datasetArtifactId": benchmark.dataset.artifact_id,
                    "parentArtifactId": parent_artifact_id,
                    **benchmark_json,
                },
            )

            inserted_benchmark = response.json()

            # We modify the slug in the server
            # Update benchmark.slug here so benchmark.urn is constructed correctly
            benchmark.slug = inserted_benchmark["slug"]

            with StorageSession(self, "write", benchmark.urn) as storage:
                logger.info("Copying the benchmark split to the Hub. This may take a while.")

                # 2. Upload each index set bitmap
                with track_progress(
                    description="Copying index sets", total=benchmark.split.n_test_sets + 1
                ) as (progress_index_sets, task_index_sets):
                    for label, index_set in benchmark.split:
                        logger.info(f"Copying index set {label} to the Hub.")
                        storage.set_file(label, index_set.serialize())
                        progress_index_sets.update(task_index_sets, advance=1, refresh=True)

            benchmark_url = urljoin(self.settings.hub_url, response.headers.get("Content-Location"))
            progress.log(
                f"[green]Your benchmark has been successfully uploaded to the Hub.\nView it here: {benchmark_url}"
            )

    def get_competition(self, artifact_id: str) -> CompetitionSpecification:
        """Load a competition from the Polaris Hub.

        Args:
            artifact_id: The artifact identifier for the competition

        Returns:
            A `CompetitionSpecification` instance, if it exists.
        """
        url = f"/v1/competition/{artifact_id}"
        response = self._base_request_to_hub(url=url, method="GET")
        response_data = response.json()

        with StorageSession(
            self, "read", CompetitionSpecification.urn_for(*artifact_id.split("/"))
        ) as storage:
            zarr_root_path = str(storage.paths.root)

        return CompetitionSpecification(zarr_root_path=zarr_root_path, **response_data)

    def submit_competition_predictions(
        self,
        competition: CompetitionSpecification,
        competition_predictions: CompetitionPredictions,
    ):
        """Submit predictions for a competition to the Polaris Hub. The Hub will evaluate them against
        the secure test set and store the result.

        Args:
            competition: The competition to evaluate the predictions for.
            competition_predictions: The predictions and associated metadata to be submitted to the Hub.
        """
        with track_progress(description="Submitting competition predictions", total=1):
            # Prepare prediction payload for submission
            prediction_json = competition_predictions.model_dump(by_alias=True, exclude_none=True)
            prediction_payload = {
                "competitionArtifactId": f"{competition.artifact_id}",
                **prediction_json,
            }

            # Submit payload to Hub
            response = self._base_request_to_hub(
                url="/v1/competition-prediction",
                method="POST",
                json=prediction_payload,
            )
            return response

    def list_models(self, limit: int = 100, offset: int = 0) -> list[str]:
        """List all available models on the Polaris Hub.

        Args:
            limit: The maximum number of models to return.
            offset: The offset from which to start returning models.

        Returns:
            A list of models names in the format `owner/model_slug`.
        """
        with track_progress(description="Fetching models", total=1):
            json_response = self._base_request_to_hub(
                url="/v2/model", method="GET", params={"limit": limit, "offset": offset}
            ).json()
            models = [model["artifactId"] for model in json_response["data"]]

            return models

    def get_model(self, artifact_id: str) -> Model:
        url = f"/v2/model/{artifact_id}"
        response = self._base_request_to_hub(url=url, method="GET")
        response_data = response.json()

        return Model(**response_data)

    def upload_model(
        self,
        model: Model,
        access: AccessType = "private",
        owner: HubOwner | str | None = None,
        parent_artifact_id: str | None = None,
    ):
        """Upload a model to the Polaris Hub.

        Info: Owner
            You have to manually specify the owner in the model data model. Because the owner could
            be a user or an organization, we cannot automatically infer this from just the logged-in user.

        Note: Required metadata
            The Polaris client and Hub maintain different requirements as to which metadata is required.
            The requirements by the Hub are stricter, so when uploading to the Hub you might
            get some errors on missing metadata. Make sure to fill-in as much of the metadata as possible
            before uploading.

        Args:
            model: The model to upload.
            access: Grant public or private access to result
            owner: Which Hub user or organization owns the artifact. Takes precedence over `model.owner`.
            parent_artifact_id: The `owner/slug` of the parent model, if uploading a new version of a model.
        """
        with track_progress(description="Uploading model", total=1) as (progress, task):
            # Get the serialized model data-structure
            model.owner = HubOwner.normalize(owner or model.owner)
            model_json = model.model_dump(by_alias=True, exclude_none=True)

            # Make a request to the Hub
            url = f"/v2/model/{model.artifact_id}"
            response = self._base_request_to_hub(
                url=url,
                method="PUT",
                json={"access": access, "parentArtifactId": parent_artifact_id, **model_json},
            )

            # NOTE: When we merge in the competition model feature, we will need to update the slug with the inserted model slug to make sure we write to the correct storage location.

            # Inform the user about where to find their newly created artifact.
            model_url = urljoin(self.settings.hub_url, response.headers.get("Content-Location"))
            progress.log(
                f"[green]Your model has been successfully uploaded to the Hub. View it here: {model_url}"
            )



import numpy as np
from fastpdb import struc
from numcodecs import MsgPack, register_codec
from numcodecs.vlen import VLenBytes
from rdkit import Chem


class RDKitMolCodec(VLenBytes):
    """
    Codec for RDKit's Molecules.

    Info: Binary strings for serialization
        This class converts the molecules to binary strings (for ML purposes, this should be lossless).
        This might not be the most storage efficient, but is fastest and easiest to maintain.
        See this [Github Discussion](https://github.com/rdkit/rdkit/discussions/7235) for more info.

    """

    codec_id = "rdkit_mol"

    def encode(self, buf: np.ndarray):
        """
        Encode a chunk of RDKit Mols to byte strings
        """
        # NOTE (cwognum): I ran into a Cython issue because we could pass None to the VLenBytes codec.
        # Using np.full() ensures all elements are initialized as empty byte strings instead.
        to_encode = np.full(fill_value=b"", shape=len(buf), dtype=object)
        for idx, mol in enumerate(buf):
            if mol is None or (isinstance(mol, bytes) and len(mol) == 0):
                continue
            if not isinstance(mol, Chem.Mol):
                raise ValueError(f"Expected an RDKitMol, but got {type(mol)} instead.")
            props = Chem.PropertyPickleOptions.AllProps
            to_encode[idx] = mol.ToBinary(props)

        to_encode = np.array(to_encode, dtype=object)
        return super().encode(to_encode)

    def decode(self, buf, out=None):
        """Decode the variable length bytes encoded data into a RDKit Mol."""
        dec = super().decode(buf, out)
        for idx, mol in enumerate(dec):
            if len(mol) == 0:
                continue
            dec[idx] = Chem.Mol(mol)

        if out is not None:
            np.copyto(out, dec)
            return out
        else:
            return dec


class AtomArrayCodec(MsgPack):
    """
    Codec for FastPDB (i.e. Biotite) Atom Arrays.

    Info: Only the most essential structural information of a protein is retained
        This conversion saves the 3D coordinates, chain ID, residue ID, insertion code, residue name, heteroatom indicator, atom name, element, atom ID, B-factor, occupancy, and charge.
        Records such as CONECT (connectivity information), ANISOU (anisotropic Temperature Factors), HETATM (heteroatoms and ligands) are handled by `fastpdb`.
        We believe this makes for a good _ML-ready_ format, but let us know if you require any other information to be saved.


    Info: PDBs as ND-arrays using `biotite`
        To save PDBs in a Polaris-compatible format, we convert them to ND-arrays using `fastpdb` and `biotite`.
        We then save these ND-arrays to Zarr archives.
        For more info, see [fastpdb](https://github.com/biotite-dev/fastpdb)
        and [biotite](https://github.com/biotite-dev/biotite/blob/main/src/biotite/structure/atoms.py)

    This codec is a subclass of the `MsgPack` codec from the `numcodecs`
    """

    codec_id = "atom_array"

    def encode(self, buf: np.ndarray):
        """
        Encode a chunk of AtomArrays to a plain Python structure that MsgPack can encode
        """

        to_pack = np.empty_like(buf)

        for idx, atom_array in enumerate(buf):
            # A chunk can have missing values
            if atom_array is None:
                continue

            if not isinstance(atom_array, struc.AtomArray):
                raise ValueError(f"Expected an AtomArray, but got {type(atom_array)} instead")

            data = {
                "coord": atom_array.coord,
                "chain_id": atom_array.chain_id,
                "res_id": atom_array.res_id,
                "ins_code": atom_array.ins_code,
                "res_name": atom_array.res_name,
                "hetero": atom_array.hetero,
                "atom_name": atom_array.atom_name,
                "element": atom_array.element,
                "atom_id": atom_array.atom_id,
                "b_factor": atom_array.b_factor,
                "occupancy": atom_array.occupancy,
                "charge": atom_array.charge,
            }
            data = {k: v.tolist() for k, v in data.items()}
            to_pack[idx] = data

        return super().encode(to_pack)

    def decode(self, buf, out=None):
        """Decode the MsgPack decoded data into a `fastpdb` AtomArray."""

        dec = super().decode(buf, out)

        structs = np.empty(shape=len(dec), dtype=object)

        for idx, data in enumerate(dec):
            if data is None:
                continue

            atom_array = []
            array_length = len(data["coord"])

            for ind in range(array_length):
                atom = struc.Atom(
                    coord=data["coord"][ind],
                    chain_id=data["chain_id"][ind],
                    res_id=data["res_id"][ind],
                    ins_code=data["ins_code"][ind],
                    res_name=data["res_name"][ind],
                    hetero=data["hetero"][ind],
                    atom_name=data["atom_name"][ind],
                    element=data["element"][ind],
                    b_factor=data["b_factor"][ind],
                    occupancy=data["occupancy"][ind],
                    charge=data["charge"][ind],
                    atom_id=data["atom_id"][ind],
                )
                atom_array.append(atom)

            # Note that this is a `fastpdb` AtomArray, not a NumPy array.
            structs[idx] = struc.array(atom_array)

        if out is not None:
            np.copyto(out, structs)
            return out
        else:
            return structs


register_codec(RDKitMolCodec)
register_codec(AtomArrayCodec)
